try:
    from imgui_bundle import imgui, implot, immapp, hello_imgui, glfw_utils
    import glfw
except ImportError:
    raise ImportError('imgui_bundle (or one of its dependencies) is not installed, GUI functionality is not available')

import threading
import numpy as np

from . import gaze_headref

class TargetPos:
    def __init__(self, timestamp: float, frame_idx: int, cam_pos: np.ndarray):
        self.timestamp = timestamp
        self.frame_idx = frame_idx
        self.cam_pos   = cam_pos

# GUI for synchronizing two signals
class GUI:
    def __init__(self, use_thread=True):
        self._should_exit = False
        self._use_thread = use_thread # NB: on MacOSX the GUI needs to be on the main thread, see https://github.com/pthom/hello_imgui/issues/33
        self._thread = None

        self.title      = ''
        self.ival       = -1
        self.gaze_data  : dict[str, np.ndarray] = {}    # ['ts', 'x', 'y']
        self.target_data: dict[str, np.ndarray] = {}
        self.offset_t   : float = 0.
        self._offset_xy : list[float] = [0., 0.]
        self._dragging  : list[bool] = [False, False]
        self._temp_off  = np.zeros((3,))
        self._should_rescale = True
        self.is_done    : bool = None

        self._window_flags = int(
                                    imgui.WindowFlags_.no_title_bar |
                                    imgui.WindowFlags_.no_collapse |
                                    imgui.WindowFlags_.no_scrollbar |
                                    imgui.WindowFlags_.no_scroll_with_mouse
                                )

    def __del__(self):
        self.stop()

    def start(self):
        if self._use_thread:
            if self._thread is not None:
                raise RuntimeError('The gui is already running, cannot start again')
            self._thread = threading.Thread(target=self._thread_start_fun)
            self._thread.start()
        else:
            self._thread_start_fun()

    def _thread_start_fun(self):
        self._should_exit = False
        self._user_closed_window = False

        def close_callback(window: glfw._GLFWwindow):
            self._user_closed_window = True

        def post_init():
            imgui.get_io().config_viewports_no_decoration = False
            imgui.get_io().config_viewports_no_auto_merge = True

            glfw.set_window_close_callback(glfw_utils.glfw_window_hello_imgui(), close_callback)

        params = hello_imgui.RunnerParams()
        params.app_window_params.window_geometry.full_screen_mode = hello_imgui.FullScreenMode.full_monitor_work_area
        params.app_window_params.restore_previous_geometry = False
        params.ini_folder_type = hello_imgui.IniFolderType.temp_folder  # so we don't have endless ini files in the app folder, since we don't use them anyway (see previous line, restore_previous_geometry = False)
        params.app_window_params.hidden = True
        params.fps_idling.fps_idle = 0
        params.callbacks.show_gui  = self._gui_func
        params.callbacks.post_init = post_init

        # multiple window support
        params.imgui_window_params.config_windows_move_from_title_bar_only = True
        params.imgui_window_params.enable_viewports = True

        addons = immapp.AddOnsParams()
        addons.with_implot = True
        immapp.run(params, addons)

    def get_state(self):
        return (self._user_closed_window, self.is_done)

    def set_data(self, title: str, ival: int, gazes: dict[int,gaze_headref.Gaze], target_positions: dict[int,TargetPos],
                 offset_t = 0.):
        if not gazes or not target_positions:
            raise ValueError('No gaze or target position data was provided')

        new_title = f'{title} (episode {ival})'
        # this is just for show, doesn't trigger an update. But lets keep them in sync
        hello_imgui.get_runner_params().app_window_params.window_title = new_title
        # actually update window title
        win = glfw_utils.glfw_window_hello_imgui()
        glfw.set_window_title(win, new_title)
        hello_imgui.get_runner_params().app_window_params.window_geometry.resize_app_window_at_next_frame = True
        hello_imgui.get_runner_params().app_window_params.hidden = False
        self.offset_t   = offset_t
        self._offset_xy = [0., 0.]
        self.is_done    = False
        self._should_rescale = True

        self.gaze_data['ts']= np.array([s.timestamp       for fr in gazes for s in gazes[fr]])/1000 # ms -> s
        self.gaze_data['x'] = np.array([s.gaze_pos_vid[0] for fr in gazes for s in gazes[fr]])
        self.gaze_data['y'] = np.array([s.gaze_pos_vid[1] for fr in gazes for s in gazes[fr]])

        self.target_data['ts']  = np.array([target_positions[fr].timestamp  for fr in target_positions])/1000 # ms -> s
        self.target_data['x']   = np.array([target_positions[fr].cam_pos[0] for fr in target_positions])
        self.target_data['y']   = np.array([target_positions[fr].cam_pos[1] for fr in target_positions])

        # set begin to 0
        t0 = min([self.gaze_data['ts'][0] , self.target_data['ts'][0] ])
        self.gaze_data  ['ts'] -= t0
        self.target_data['ts'] -= t0

    def _gui_func(self):
        # check if we should exit
        if self._should_exit:
            # and kill
            hello_imgui.get_runner_params().app_shall_exit = True
            # nothing more to do
            return

        # check we have data to plot
        if not self.gaze_data:
            return

        if imgui.button('Done', (-1,0)):
            self.is_done = True

        toff = self.offset_t
        xoff = self._offset_xy[0]
        yoff = self._offset_xy[1]
        if any(self._dragging):
            toff += self._temp_off[0]
            xoff += self._temp_off[1]
            yoff += self._temp_off[2]
        if implot.begin_subplots('##xy_plots',2,1,(-1,-1),implot.SubplotFlags_.link_all_x)[0]:
            if self._should_rescale:
                implot.set_next_axes_to_fit()
            if implot.begin_plot('##X',flags=implot.Flags_.no_mouse_text):
                implot.setup_axis(implot.ImAxis_.x1, None)
                implot.setup_axis(implot.ImAxis_.y1, 'horizontal coordinate (pix)')
                implot.plot_line("gaze", self.gaze_data['ts']+toff, self.gaze_data['x']+xoff)
                implot.plot_line("target", self.target_data['ts'], self.target_data['x'])
                self._do_drag(0)
                # position annotation outside axes, clamping will put it in the corner
                ax_lims = implot.get_plot_limits()
                implot.annotation(ax_lims.x.max,ax_lims.y.min,(0,0,0,0),(0,0),True,f'offset: {toff*1000:.1f}ms')
                implot.end_plot()

            if self._should_rescale:
                implot.set_next_axes_to_fit()
                self._should_rescale = False
            if implot.begin_plot('##Y',flags=implot.Flags_.no_mouse_text):
                implot.setup_axis(implot.ImAxis_.x1, 'time (s)')
                implot.setup_axis(implot.ImAxis_.y1, 'vertical coordinate (pix)')
                implot.plot_line("gaze", self.gaze_data['ts']+toff, self.gaze_data['y']+yoff)
                implot.plot_line("target", self.target_data['ts'], self.target_data['y'])
                self._do_drag(1)
                implot.end_plot()

            implot.end_subplots()

    def _do_drag(self, ax_idx):
        ax_lims = implot.get_plot_limits()
        pos = ((ax_lims.x.min+ax_lims.x.max)/2., (ax_lims.y.min+ax_lims.y.max)/2.)
        held, dx, dy = implot.drag_point(1,*pos,imgui.ImVec4(0,0.9,0,1))[:3]
        if held:
            self._dragging[ax_idx]  = True
            self._temp_off[0]       = dx-pos[0]
            self._temp_off[ax_idx+1]= dy-pos[1]
        elif self._dragging[ax_idx]:
            self.offset_t           += self._temp_off[0]
            self._offset_xy[ax_idx] += self._temp_off[ax_idx+1]
            self._temp_off          = np.zeros((3,))
            self._dragging[ax_idx]  = False

    def stop(self):
        self._should_exit = True
        if self._thread is not None:
            self._thread.join()
        self._thread = None


import numpy as np
import math
import cv2

from . import gaze_headref, gaze_worldref, marker, ocv, plane

def toNormPos(x,y,bbox):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a normalized position
    # in an image of the plane, given the image's bounding box in
    # world units
    # for input (0,0) is bottom left, for output (0,0) is top left
    # bbox is [left, top, right, bottom]

    extents = [bbox[2]-bbox[0], bbox[1]-bbox[3]]
    pos     = [(x-bbox[0])/extents[0], (bbox[1]-y)/extents[1]]    # bbox[1]-y instead of y-bbox[3] to flip y
    return pos

def toImagePos(x,y,bbox,imSize,margin=[0,0]):
    # transforms input (x,y) which is on a plane in world units
    # (e.g. mm on an aruco poster) to a pixel position in the
    # image, given the image's bounding box in world units
    # imSize should be active image area in pixels, excluding margin

    # fractional position between bounding box edges, (0,0) in bottom left
    pos = toNormPos(x,y, bbox)
    # turn into int, add margin
    pos = [p*s+m for p,s,m in zip(pos,imSize,margin)]
    return pos


def estimateHomographyKnownMarker(known: list[marker.Marker], detectedCorners, detectedIDs):
    # collect matching corners in image and in world
    imgP = []
    objP = []
    detectedIDs = detectedIDs.flatten()
    if len(detectedIDs) != len(detectedCorners):
        raise ValueError('unexpected number of IDs (%d) given number of corner arrays (%d)' % (len(detectedIDs),len(detectedCorners)))
    for i in range(0, len(detectedIDs)):
        if detectedIDs[i] in known:
            dc = detectedCorners[i]
            if dc.shape[0]==1 and dc.shape[1]==4:
                dc = np.reshape(dc,(4,1,2))
            imgP.extend( [x.flatten() for x in dc] )
            objP.extend( known[detectedIDs[i]].corners )

    if len(imgP) < 4:
        return None, False

    # compute Homography
    return estimateHomography(objP, imgP)

def estimateHomography(objP, imgP):
    imgP = np.float32(imgP)
    objP = np.float32(objP)
    h, _ = cv2.findHomography(imgP, objP)
    return h, True

def applyHomography(H, x, y):
    if math.isnan(x) or math.isnan(y):
        return np.full((2,), np.nan)

    src = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    dst = cv2.perspectiveTransform(src,H)
    return dst.flatten()


def distortPoint(x, y, cameraMatrix, distCoeff):
    if math.isnan(x) or math.isnan(y):
        return np.full((2,), np.nan)

    # unproject, ignoring distortion as this is an undistored point
    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, cameraMatrix, np.asarray([[0.0, 0.0, 0.0, 0.0, 0.0]]))
    points_3d = cv2.convertPointsToHomogeneous(points_2d)
    points_3d.shape = -1, 3

    # reproject, applying distortion
    points_2d, _ = cv2.projectPoints(points_3d, np.zeros((1, 1, 3)), np.zeros((1, 1, 3)), cameraMatrix, distCoeff)
    return points_2d.flatten()

def undistortPoint(x, y, cameraMatrix, distCoeff):
    if math.isnan(x) or math.isnan(y):
        return np.full((2,), np.nan)

    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, cameraMatrix, distCoeff, P=cameraMatrix) # P=cameraMatrix to reproject to camera
    return points_2d.flatten()

def unprojectPoint(x, y, cameraMatrix, distCoeff):
    if math.isnan(x) or math.isnan(y):
        return np.full((3,), np.nan)

    points_2d = np.asarray([x, y], dtype='float32').reshape((1, -1, 2))
    points_2d = cv2.undistortPoints(points_2d, cameraMatrix, distCoeff)
    points_3d = cv2.convertPointsToHomogeneous(points_2d)
    points_3d.shape = -1, 3
    return points_3d.flatten()


def intersect_plane_ray(planeNormal, planePoint, rayDirection, rayPoint, epsilon=1e-6):
    # from https://rosettacode.org/wiki/Find_the_intersection_of_a_line_with_a_plane#Python

    ndotu = planeNormal.dot(rayDirection)
    if abs(ndotu) < epsilon:
        # raise RuntimeError("no intersection or line is within plane")
        return np.full((3,), np.nan)

    w = rayPoint - planePoint
    si = -planeNormal.dot(w) / ndotu
    return w + si * rayDirection + planePoint


def angle_between(v1, v2):
    return (180.0 / math.pi) * math.atan2(np.linalg.norm(np.cross(v1,v2)), np.dot(v1,v2))


def gazeToPlane(gaze: gaze_headref.Gaze, pose: plane.Pose, cameraParams: ocv.CameraParams) -> gaze_worldref.Gaze:
    gazeWorld     = gaze_worldref.Gaze(gaze.timestamp, gaze.frame_idx)
    if pose.pose_N_markers>0:
        # get transform from ET data's coordinate frame to camera's coordinate frame
        cameraRotation = cameraParams.rotation_vec
        cameraPosition = cameraParams.position
        if cameraRotation is None:
            cameraRotation = np.zeros((3,1))
        RCam  = cv2.Rodrigues(cameraRotation)[0]
        if cameraPosition is None:
            cameraPosition = np.zeros((3,1))
        RtCam = np.hstack((RCam, cameraPosition))

        # project gaze on video to reference poster using camera pose
        gazeWorld.gazePosPlane2D_vidPos_ray, gazeWorld.gazePosCam_vidPos_ray = \
            pose.camToPlanePose(gaze.gaze_pos_vid, cameraParams)

        # project world-space gaze point (often binocular gaze point) to plane
        if gaze.gaze_pos_3d is not None:
            # transform 3D gaze point from eye tracker space to camera space
            g3D = np.matmul(RtCam,np.array(np.append(gaze.gaze_pos_3d, 1)).reshape(4,1))

            # find intersection with poster (NB: pose is in camera reference frame)
            gazeWorld.gazePosCamWorld = pose.vectorIntersect(g3D)    # default vec origin (0,0,0) is fine because we work from camera's view point

            # above intersection is in camera space, turn into poster space to get position on poster
            (x,y,z) = pose.camToWorld(gazeWorld.gazePosCamWorld)   # z should be very close to zero
            gazeWorld.gazePosPlane2DWorld = np.asarray([x, y])

    # unproject 2D gaze point on video to point on poster (should yield values very close to
    # the above method of intersecting video gaze point ray with poster, and usually also very
    # close to binocular gaze point (though for at least one tracker the latter is not the case;
    # the AdHawk has an optional parallax correction using a vergence signal))
    if pose.homography_N_markers>0:
        gazeWorld.gazePosPlane2D_vidPos_homography = pose.camToPlaneHomography(gaze.gaze_pos_vid, cameraParams)

        # get this point in camera space
        if pose.pose_N_markers>0:
            gazeWorld.gazePosCam_vidPos_homography = pose.worldToCam(np.append(gazeWorld.gazePosPlane2D_vidPos_homography, 0))

    # project gaze vectors to reference poster (and draw on video)
    if not pose.pose_N_markers>0:
        # nothing to do anymore
        return gazeWorld

    gazeVecs    = [gaze.gaze_dir_l, gaze.gaze_dir_r]
    gazeOrigins = [gaze.gaze_ori_l, gaze.gaze_ori_r]
    attrs       = [['gazeOriCamLeft','gazePosCamLeft','gazePosPlane2DLeft'],['gazeOriCamRight','gazePosCamRight','gazePosPlane2DRight']]
    for gVec,gOri,attr in zip(gazeVecs,gazeOrigins,attrs):
        if gVec is None or gOri is None:
            continue
        # get gaze vector and point on vector (origin, e.g. pupil center) ->
        # transform from ET data coordinate frame into camera coordinate frame
        gVec    = np.matmul(RCam ,          gVec    )
        gOri    = np.matmul(RtCam,np.append(gOri,1.))
        setattr(gazeWorld,attr[0],gOri)

        # intersect with poster -> yield point on poster in camera reference frame
        gPoster = pose.vectorIntersect(gVec, gOri)
        setattr(gazeWorld,attr[1],gPoster)

        # transform intersection with poster from camera space to poster space
        (x,y,z)  = pose.camToWorld(gPoster)  # z should be very close to zero
        setattr(gazeWorld,attr[2],np.asarray([x, y]))

    return gazeWorld


import numpy as np
import cv2
from typing import Any

from . import drawing, intervals, marker, ocv, plane, timestamps



class ArUcoDetector():
    def __init__(self, ArUco_dict: cv2.aruco.Dictionary, params: dict[str, Any]):
        self._det_params                          = cv2.aruco.DetectorParameters()
        self._det_params.cornerRefinementMethod   = cv2.aruco.CORNER_REFINE_SUBPIX    # good default, user can override
        for p in params:
            setattr(self._det_params, p, params[p])

        self._dict                      = ArUco_dict
        self._det                       = cv2.aruco.ArucoDetector(self._dict, self._det_params)
        self._board: cv2.aruco.Board    = None

        self._camera_params             = ocv.CameraParams(None, None)

    def create_board(self, board_corner_points, ids):
        self._board = create_board(board_corner_points, ids, self._dict)

    def set_board(self, board):
        self._board = board

    def set_intrinsics(self, camera_params: ocv.CameraParams):
        self._camera_params = camera_params

    def detect_markers(self, image: cv2.UMat, min_nmarker_refine = 3):
        corners, ids, rejected_corners = _detect_markers(image, self._det)
        recovered_ids = None

        # Refine detected markers (eliminates markers not part of our poster, adds missing markers to the poster)
        if self._board and ids is not None and min_nmarker_refine and len(ids)>min_nmarker_refine:
            corners, ids, rejected_corners, recovered_ids = \
                self.refine_detected_markers(image, corners, ids, rejected_corners)
        return corners, ids, rejected_corners, recovered_ids

    def refine_detected_markers(self, image, detected_corners, detected_ids, rejected_corners):
        return _refine_detection(image, detected_corners, detected_ids, rejected_corners, self._det, self._board, self._camera_params.camera_mtx, self._camera_params.distort_coeffs)

    def _match_image_points(self, corners, ids):
        return self._board.matchImagePoints(corners, ids) # -> objP, imgP

    def estimate_pose(self, corners, ids) -> tuple[int, np.ndarray, np.ndarray]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_pose_impl(objP, imgP)

    def _estimate_pose_impl(self, objP, imgP):
        N_markers, R_vec, T_vec = 0, None, None
        if objP is None or not self._camera_params.has_intrinsics():
            return N_markers, R_vec, T_vec

        ok, R_vec, T_vec = cv2.solvePnP(objP, imgP, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, np.empty(1), np.empty(1))
        if ok:
            N_markers = int(objP.shape[0]/4)
        return N_markers, R_vec, T_vec

    def estimate_homography(self, corners, ids) -> tuple[np.ndarray, bool]:
        objP, imgP = self._match_image_points(corners, ids)
        return self._estimate_homography_impl(objP, imgP)

    def _estimate_homography_impl(self, objP, imgP):
        from . import transforms
        N_markers, H = 0, None
        if objP is None:
            return N_markers, H

        # use undistorted marker corners if possible
        if self._camera_params.has_intrinsics():
            imgP = np.vstack([cv2.undistortPoints(x, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, P=self._camera_params.camera_mtx) for x in imgP])

        H, status = transforms.estimateHomography(objP, imgP)
        if status:
            N_markers = int(objP.shape[0]/4)
        return N_markers, H

    # higher level functions for detecting + pose estimation, and for results visualization
    def estimate_pose_homography(self, frame_idx, min_num_markers, corners, ids) -> tuple[plane.Pose, dict[str,Any]]:
        pose = plane.Pose(frame_idx)
        if ids is not None and len(ids) >= min_num_markers:
            # get matching image and board points
            objP, imgP = self._match_image_points(corners, ids)

            # get camera pose
            pose.pose_N_markers, pose.pose_R_vec, pose.pose_T_vec = \
                self._estimate_pose_impl(objP, imgP)

            # also get homography (direct image plane to plane in world transform)
            pose.homography_N_markers, pose.homography_mat = \
                self._estimate_homography_impl(objP, imgP)
        return pose

    def detect_and_estimate(self, frame, frame_idx, min_num_markers) -> tuple[plane.Pose, dict[str,Any]]:
        corners, ids, rejectedImgPoints, recoveredIds = self.detect_markers(frame, min_nmarker_refine=min_num_markers)
        pose = self.estimate_pose_homography(frame_idx, min_num_markers, corners, ids)
        return pose, {'corners': corners, 'ids': ids, 'rejectedImgPoints': rejectedImgPoints, 'recoveredIds': recoveredIds}

    def visualize(self, frame, pose: plane.Pose, detect_dict, arm_length, sub_pixel_fac = 8, show_rejected_markers = False):
        if pose.pose_N_markers>0:
            # draw axis indicating poster pose (origin and orientation)
            drawing.openCVFrameAxis(frame, self._camera_params.camera_mtx, self._camera_params.distort_coeffs, pose.pose_R_vec, pose.pose_T_vec, arm_length, 3, sub_pixel_fac)

        if pose.homography_N_markers>0:
            # find where plane origin is expected to be in the image
            target = pose.planeToCamHomography([0., 0.], self._camera_params)
            # draw target location on image
            if target[0] >= 0 and target[0] < frame.shape[1] and target[1] >= 0 and target[1] < frame.shape[0]:
                drawing.openCVCircle(frame, target, 3, (0,0,0), -1, sub_pixel_fac)

        # if any markers were detected, draw where on the frame
        drawing.arucoDetectedMarkers(frame, detect_dict['corners'], detect_dict['ids'], subPixelFac=sub_pixel_fac, specialHighlight=[detect_dict['recoveredIds'],(255,255,0)])

        # for debug, can draw rejected markers on frame
        if show_rejected_markers:
            cv2.aruco.drawDetectedMarkers(frame, detect_dict['rejectedImgPoints'], None, borderColor=(211,0,148))



def _detect_markers(image: cv2.UMat, det: cv2.aruco.ArucoDetector):
    corners, ids, rejected_corners = det.detectMarkers(image)
    if np.any(ids==None):
        ids = None
    return corners, ids, rejected_corners

def _refine_detection(image: cv2.UMat, detected_corners, detected_ids, rejected_corners, det: cv2.aruco.ArucoDetector, board: cv2.aruco.Board, camera_mtx, distort_coeffs):
    corners, ids, rejectedImgPoints, recoveredIds = \
        det.refineDetectedMarkers(
            image = image, board = board,
            detectedCorners = detected_corners, detectedIds = detected_ids, rejectedCorners = rejected_corners,
            cameraMatrix = camera_mtx, distCoeffs = distort_coeffs
            )
    if corners and corners[0].shape[0]==4:
        # there are versions out there where there is a bug in output shape of each set of corners, fix up
        corners = [np.reshape(c,(1,4,2)) for c in corners]
    if rejectedImgPoints and rejectedImgPoints[0].shape[0]==4:
        # same as for corners
        rejectedImgPoints = [np.reshape(c,(1,4,2)) for c in rejectedImgPoints]

    return corners, ids, rejectedImgPoints, recoveredIds

def _refine_for_multiple_planes(image, detected_corners, detected_ids, rejected_corners, detectors: dict[str, ArUcoDetector], camera_mtx, distort_coeffs):
    result = {}
    for d in detectors:
        a,b,c,e = _refine_detection(image, detected_corners, detected_ids, rejected_corners, detectors[d]._det, detectors[d]._board, detectors[d]._camera_params.camera_mtx, detectors[d]._camera_params.distort_coeffs)
        result[d] = dict(zip(['corners', 'ids', 'rejectedImgPoints', 'recoveredIds'],(a,b,c,e)))
    return result

def create_board(board_corner_points: list[np.ndarray], ids: list[int], ArUco_dict: cv2.aruco.Dictionary):
    board_corner_points = np.dstack(board_corner_points)        # list of 2D arrays -> 3D array
    board_corner_points = np.rollaxis(board_corner_points,-1)   # 4x2xN -> Nx4x2
    board_corner_points = np.pad(board_corner_points,((0,0),(0,0),(0,1)),'constant', constant_values=(0.,0.)) # Nx4x2 -> Nx4x3
    return cv2.aruco.Board(board_corner_points, ArUco_dict, np.array(ids))

def run_pose_estimation(in_video, frame_timestamp_file, camera_calibration_file,
                        output_dir,
                        processing_intervals,
                        planes, individual_markers,
                        gui, sub_pixel_fac = 8, show_rejected_markers = False) -> tuple[bool, dict[str,list[plane.Pose]], dict[str,list[marker.Pose]]]:
    show_visualization = gui is not None

    # open video
    cap = ocv.CV2VideoReader(in_video, timestamps.VideoTimestamps(frame_timestamp_file).timestamps)

    # setup aruco marker detection
    aruco_boards = {p: planes[p]['plane'].get_aruco_board() for p in planes}
    detectors = {p: ArUcoDetector(aruco_boards[p].getDictionary(), planes[p]['aruco_params']) for p in planes}
    cam_params = ocv.CameraParams.readFromFile(camera_calibration_file)
    for p in detectors:
        detectors[p].set_board(aruco_boards[p])
        # get camera calibration info
        detectors[p].set_intrinsics(cam_params)

    # check if we can do an optimization of detecting the markers only once for multiple planes (if it makes sense because we have more than one plane)
    plane_names = list(planes.keys())
    aruco_dicts = [planes[p]['aruco_dict'] for p in planes if 'aruco_dict' in planes[p]]
    all_same_dict_and_dect = len(planes)>1 and len(aruco_dicts)==len(planes) and len(set(aruco_dicts))==1 and all([planes[plane_names[0]]['aruco_params']==planes[p]['aruco_params'] for p in planes])

    # check individual markers (detected using the same detector as for the plane(s))
    has_individual_markers = individual_markers is not None
    if has_individual_markers:
        assert all_same_dict_and_dect or len(planes)==1, "Detecting and Reporting individual markers are only supported when there is a single plane, or all planes have identical aruco setup"
        individual_markers_out = {i:[] for i in individual_markers}
        object_points = {}
        for i in individual_markers:
            marker_size = individual_markers[i]['marker_size']
            object_points[i] = np.array([[-marker_size/2, marker_size/2, 0],[marker_size/2, marker_size/2, 0],[marker_size/2, -marker_size/2, 0],[-marker_size/2, -marker_size/2, 0]])
    else:
        individual_markers_out = None

    stop_all_processing = False
    poses = {p:[] for p in planes}
    while True:
        # process frame-by-frame
        done, frame, frame_idx, frame_ts = cap.read_frame(report_gap=True)
        if done or intervals.beyond_last_interval(frame_idx, processing_intervals):
            break
        cap.report_frame()

        if frame is None:
            # we don't have a valid frame, continue to next
            continue

        if show_visualization:
            keys = gui.get_key_presses()
            if 'q' in keys:
                # quit fully
                stop_all_processing = True
                break
            if 'n' in keys:
                # goto next
                break

        # check we're in a current interval, else skip processing
        # NB: have to spool through like this, setting specific frame to read
        # with cap.set(cv2.CAP_PROP_POS_FRAMES) doesn't seem to work reliably
        # for VFR video files
        planes_for_this_frame = [p for p in planes if intervals.is_in_interval(frame_idx, processing_intervals[p])]
        if not planes_for_this_frame:
            continue

        # detect markers
        detect_dicts = {}
        if all_same_dict_and_dect or has_individual_markers:
            corners, ids, rejected_corners = _detect_markers(frame, detectors[plane_names[0]]._det)
            detect_dicts = _refine_for_multiple_planes(frame, corners, ids, rejected_corners, {p:detectors[p] for p in planes_for_this_frame}, cam_params.camera_mtx, cam_params.distort_coeffs)
        else:
            for p in planes_for_this_frame:
                detect_dicts[p] = dict(zip(['corners', 'ids', 'rejectedImgPoints', 'recoveredIds'], detectors[p].detect_markers(frame, planes[p]['min_num_markers'])))
        # determine pose
        for p in planes_for_this_frame:
            pose = detectors[p].estimate_pose_homography(frame_idx, planes[p]['min_num_markers'], detect_dicts[p]['corners'], detect_dicts[p]['ids'])
            poses[p].append(pose)
            # draw detection and pose, if wanted
            if show_visualization:
                detectors[p].visualize(frame, pose, detect_dicts[p], planes[p]['plane'].marker_size/2, sub_pixel_fac, show_rejected_markers)

        # deal with individual markers, if any
        if has_individual_markers and ids is not None:
            found_markers = np.where([x[0] in individual_markers for x in ids])[0]
            if found_markers.size>0:
                for idx in found_markers:
                    m_id = ids[idx][0]
                    pose = marker.Pose(frame_idx)
                    ret, pose.R_vec, pose.T_vec = cv2.solvePnP(object_points[m_id], corners[idx], cam_params.camera_mtx, cam_params.distort_coeffs, flags=cv2.SOLVEPNP_IPPE_SQUARE)
                    if not ret:
                        continue
                    individual_markers_out[m_id].append(pose)
                    if show_visualization:
                        pose.draw_origin_on_frame(frame, cam_params.camera_mtx, cam_params.distort_coeffs, individual_markers[m_id]['marker_size']/2, sub_pixel_fac)

        if show_visualization:
            # keys is populated above
            if 's' in keys:
                # screenshot
                cv2.imwrite(output_dir / f'detect_frame_{frame_idx}.png', frame)
            gui.update_image(frame, frame_ts/1000., frame_idx)
            closed, = gui.get_state()
            if closed:
                stop_all_processing = True
                break

    if show_visualization:
        gui.stop()

    return stop_all_processing, poses, individual_markers_out


"""
Cast raw SeeTrue data into common format.

The output directory will contain:
    - frameTimestamps.tsv: frame number and corresponding timestamps for each frame in video
    - worldCamera.mp4: the video from the point-of-view scene camera on the glasses
    - gazeData.tsv: gaze data, where all 2D gaze coordinates are represented w/r/t the world camera
"""

import shutil
import pathlib
import os
import cv2
import pandas as pd
import numpy as np

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import video_utils

def preprocessData(output_dir: str|pathlib.Path, source_dir: str|pathlib.Path=None, rec_info: Recording=None, cam_cal_file: str|pathlib.Path=None, copy_scene_video = True) -> Recording:
    from . import check_folders
    # NB: copy_scene_video input argument is ignored, SeeTrue recordings must be transcoded with ffmpeg to be useful

    if shutil.which('ffmpeg') is None:
        raise RuntimeError('ffmpeg not found on path. ffmpeg is required for importing SeeTrue recordings. Cannot continue')

    """
    Run all preprocessing steps on SeeTrue STONE data and store in output_dir
    """
    output_dir, source_dir, rec_info = check_folders(output_dir, source_dir, rec_info, EyeTracker.SeeTrue_STONE)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    if rec_info is not None:
        if not checkRecording(source_dir, rec_info):
            raise ValueError(f"A recording with the name \"{rec_info.name}\" was not found in the folder {source_dir}.")
    else:
        recInfos = getRecordingInfo(source_dir)
        if recInfos is None:
            raise RuntimeError(f"The folder {source_dir} does not contain SeeTrue STONE recordings.")
        rec_info = recInfos[0]  # take first, arbitrarily. If anything else wanted, user should call this function with a correct rec_info themselves

    # make output dirs
    if not output_dir.is_dir():
        output_dir.mkdir()


    #### prep the data
    # NB: gaze data and scene video prep are intertwined, status messages are output inside this function
    gazeDf, frameTimestamps = copySeeTrueRecording(source_dir, output_dir, rec_info)

    print('  Getting camera calibration...')
    if cam_cal_file is not None:
        shutil.copyfile(str(cam_cal_file), str(output_dir / 'calibration.xml'))
    else:
        print('    !! No camera calibration provided! Defaulting to hardcoded')
        getCameraHardcoded(output_dir)


    # write the gaze data to a csv file
    gazeDf.to_csv(str(output_dir / 'gazeData.tsv'), sep='\t', na_rep='nan', float_format="%.8f")

    # also store frame timestamps
    frameTimestamps.to_csv(str(output_dir / 'frameTimestamps.tsv'), sep='\t')

    # store rec info
    rec_info.store_as_json(output_dir)

    return rec_info


def getRecordingInfo(inputDir: str|pathlib.Path) -> list[Recording]:
    # returns None if not a recording directory
    recInfos = []

    # NB: a SeeTrue directory may contain multiple recordings

    # get recordings. These are indicated by the sequence number in both EyeData.csv and ScenePics folder names
    for r in inputDir.glob('*.csv'):
        if not str(r.name).startswith('EyeData'):
            # print(f"file {r.name} not recognized as a recording (wrong name, should start with 'EyeData'), skipping")
            continue

        # get sequence number
        _,recording = r.stem.split('_')

        # check there is a matching scenevideo
        sceneVidDir = r.parent / ('ScenePics_' + recording)
        if not sceneVidDir.is_dir():
            # print(f"folder {sceneVidDir} not found, meaning there is no scene video for this recording, skipping")
            continue

        recInfos.append(Recording(source_directory=inputDir, eye_tracker=EyeTracker.SeeTrue_STONE))
        recInfos[-1].participant = inputDir.name
        recInfos[-1].name = recording

    # should return None if no valid recordings found
    return recInfos if recInfos else None


def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    """
    This checks that the folder is properly prepared
    (i.e. the required BeGaze exports were run)
    """
    # check we have an exported gaze data file
    file = f'EyeData_{recInfo.name}.csv'
    if not (inputDir / file).is_file():
        return False

    # check we have an exported scene video
    file = f'ScenePics_{recInfo.name}'
    if not (inputDir / file).is_dir():
        return False

    return True


def copySeeTrueRecording(inputDir: pathlib.Path, outputDir: pathlib.Path, recInfo: Recording):
    """
    Copy the relevant files from the specified input dir to the specified output dirs
    """

    # get scene video dimensions by interrogating a frame in sceneVidDir
    sceneVidDir = inputDir / ('ScenePics_' + recInfo.name)
    frame = next(sceneVidDir.glob('*.jpeg'))
    h,w,_ = cv2.imread(frame).shape

    # prep gaze data and get video frame timestamps from it
    print('  Prepping gaze data...')
    file = f'EyeData_{recInfo.name}.csv'
    gazeDf, frameTimestamps = formatGazeData(inputDir / file, [w,h])

    # make scene video
    print('  Prepping scene video...')
    # 1. see if there are frames missing, insert black frames if so
    frames = []
    for f in sceneVidDir.glob('*.jpeg'):
        _,fr = f.stem.split('_')
        frames.append(int(fr))
    frames = sorted(frames)

    # 2. see if framenumbers are as expected from the gaze data file
    # get average ifi
    ifi = np.mean(np.diff(frameTimestamps.index))
    # 2.1 remove frame timestamps that are before the first frame for which we have an image
    frameTimestamps=frameTimestamps.drop(frameTimestamps[frameTimestamps.frame_idx < frames[ 0]].index)
    # 2.2 remove frame timestamps that are beyond last frame for which we have an image
    frameTimestamps=frameTimestamps.drop(frameTimestamps[frameTimestamps.frame_idx > frames[-1]].index)
    # 2.3 add frame timestamps for images we have before first eye data
    if frames[ 0] < frameTimestamps.iloc[ 0,:].to_numpy()[0]:
        nFrames = frameTimestamps.iloc[ 0,:].to_numpy()[0] - frames[ 0]
        t0 = frameTimestamps.index[0]
        f0 = frameTimestamps.iloc[ 0,:].to_numpy()[0]
        for f in range(-1,-(nFrames+1),-1):
            frameTimestamps.loc[t0+f*ifi] = f0+f
        frameTimestamps = frameTimestamps.sort_index()
    # 2.4 add frame timestamps for images we have after last eye data
    if frames[-1] > frameTimestamps.iloc[-1,:].to_numpy()[0]:
        nFrames = frames[-1] - frameTimestamps.iloc[-1,:].to_numpy()[0]
        t0 = frameTimestamps.index[-1]
        f0 = frameTimestamps.iloc[-1,:].to_numpy()[0]
        for f in range(1,nFrames+1):
            frameTimestamps.loc[t0+f*ifi] = f0+f
        frameTimestamps = frameTimestamps.sort_index()
    # 2.5 check if holes, fill
    blackFrames = []
    frameDelta = np.diff(frames)
    if np.any(frameDelta>1):
        # frames images missing, add them (NB: if timestamps also missing, thats dealt with below)
        idxGaps = np.argwhere(frameDelta>1).flatten()     # idxGaps is last idx before each gap
        frGaps  = np.array(frames)[idxGaps].flatten()
        nFrames = frameDelta[idxGaps].flatten()
        for b,x in zip(frGaps+1,nFrames):
            for y in range(x-1):
                blackFrames.append(b+y)

        # make black frame
        blackIm = np.zeros((h,w,3), np.uint8)   # black image
        for f in blackFrames:
            # store black frame to file
            cv2.imwrite(sceneVidDir / f'frame_{f:d}.jpeg',blackIm)
            frames.append(f)
        frames = sorted(frames)

    # 3. make into video
    framerate = "{:.4f}".format(1000./ifi)
    outFile = outputDir / 'worldCamera.mp4'
    cmd_str = ' '.join(['ffmpeg', '-hide_banner', '-loglevel', 'error', '-y', '-f', 'image2', '-framerate', framerate, '-start_number', str(frames[0]), '-i', '"'+str(sceneVidDir / 'frame_%d.jpeg')+'"', '"'+str(outFile)+'"'])
    os.system(cmd_str)
    if outFile.is_file():
        recInfo.scene_video_file = outFile.name
    else:
        raise RuntimeError('Error making a scene video out of the SeeTrue''s frames')

    # attempt 2 that should allow correct VFR video files, but doesn't work with current MediaWriter
    # due to what i think is a bug: https://github.com/matham/ffpyplayer/issues/129.
    ## get which pixel format
    #codec    = ffpyplayer.tools.get_format_codec(fmt='mp4')
    #pix_fmt  = ffpyplayer.tools.get_best_pix_fmt('bgr24',ffpyplayer.tools.get_supported_pixfmts(codec))
    #fpsFrac  = Fraction(1000./ifi).limit_denominator(10000).as_integer_ratio()
    #fpsFrac  = tuple([x*10 for x in fpsFrac])
    ## scene video
    #out_opts = {'pix_fmt_in':'bgr24', 'pix_fmt_out':pix_fmt, 'width_in':w, 'height_in':h,'frame_rate':fpsFrac}
    #vidOut   = MediaWriter(str(outputDir / 'worldCamera.mp4'), [out_opts], overwrite=True)
    #t0       = frameTimestamps.index[0]
    #for i,f in enumerate(frames):
    #    frame = cv2.imread(str(sceneVidDir / 'frame_{:5d}.jpeg'.format(f)))
    #    img   = Image(plane_buffers=[frame.flatten().tobytes()], pix_fmt='bgr24', size=(int(w), int(h)))
    #    t = (frameTimestamps.index[i]-t0)/1000
    #    print(t, t/(fpsFrac[1]/fpsFrac[0]))
    #    vidOut.write_frame(img=img, pts=t)

    # delete the black frames we added, if any
    for f in blackFrames:
        if (sceneVidDir / 'frame_{:d}.jpeg'.format(f)).is_file():
            (sceneVidDir / 'frame_{:d}.jpeg'.format(f)).unlink(missing_ok=True)

    # 4. write data to file
    # fix up frame idxs and timestamps
    firstFrame = frameTimestamps['frame_idx'].min()

    # write the gaze data to a csv file
    gazeDf['frame_idx'] -= firstFrame

    # also store frame timestamps
    # this is what it should be if we get VFR video file writing above to work
    #frameTimestamps['frame_idx'] -= firstFrame
    #frameTimestamps=frameTimestamps.reset_index().set_index('frame_idx')
    #frameTimestamps['timestamp'] -= frameTimestamps['timestamp'].min()
    # instead now, get actual ts for each frame in written video as that is what we
    # have to work with. Note that these do not match gaze data ts, but code nowhere
    # assumes they do
    frameTimestamps = video_utils.getFrameTimestampsFromVideo(recInfo.get_scene_video_path())

    return gazeDf, frameTimestamps


def getCameraHardcoded(outputDir: str|pathlib.Path):
    """
    Get camera calibration
    Hardcoded as per info received from SeeTrue
    """
    # turn into camera matrix and distortion coefficients as used by OpenCV
    camera = {}
    camera['cameraMatrix'] = np.array([[495,   0, 300],
                                       [  0, 495, 255],
                                       [  0,   0,   1]], dtype=np.float64)
    camera['distCoeff'] = np.array([-0.55, 0.4, 0, 0, -0.2])
    camera['resolution'] = np.array([640, 480])

    # store to file
    fs = cv2.FileStorage(outputDir / 'calibration.xml', cv2.FILE_STORAGE_WRITE)
    for key,value in camera.items():
        fs.write(name=key,val=value)
    fs.release()


def formatGazeData(inputFile: str|pathlib.Path, sceneVideoDimensions: list[int]):
    """
    load gazedata file
    format to get the gaze coordinates w.r.t. world camera, and timestamps for
    every frame of video

    Returns:
        - formatted dataframe with cols for timestamp, frame_idx, and gaze data
        - np array of frame timestamps
    """

    # convert the json file to pandas dataframe
    df = gazedata2df(inputFile, sceneVideoDimensions)

    # get time stamps for scene picture numbers
    frameTimestamps = pd.DataFrame(df['frame_idx'])

    # return the gaze data df and frame time stamps array
    return df, frameTimestamps


def gazedata2df(textFile: str|pathlib.Path, sceneVideoDimensions: list[int]):
    """
    convert the gazedata file to a pandas dataframe
    """

    df = pd.read_table(textFile,sep=';',index_col=False)
    df.columns=df.columns.str.strip()

    # remove unneeded columns
    rem = [x for x in df.columns if x not in ['Frame number','Timestamp','Gazepoint X','Gazepoint Y','Scene picture number']]
    df = df.drop(columns=rem)

    # rename and reorder columns
    lookup = {'Timestamp': 'timestamp',
              'Scene picture number': 'frame_idx',
              'Gazepoint X': 'gaze_pos_vid_x',
              'Gazepoint Y': 'gaze_pos_vid_y',}
    df=df.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in df.columns]
    idx.extend([x for x in df.columns if x not in idx])   # append columns not in lookup
    df = df[idx]

    # set timestamps as index
    df = df.set_index('timestamp')

    # turn gaze locations into pixel data with origin in top-left
    df['gaze_pos_vid_x'] *= sceneVideoDimensions[0]
    df['gaze_pos_vid_y'] *= sceneVideoDimensions[1]

    # return the dataframe
    return df


"""
Cast raw pupil labs (Core, Invisible and Neon) data into common format.

The output directory will contain:
    - frameTimestamps.tsv: frame number and corresponding timestamps for each frame in video
    - worldCamera.mp4: the video from the point-of-view scene camera on the glasses
    - gazeData.tsv: gaze data, where all 2D gaze coordinates are represented w/r/t the world camera,
                    and 3D coordinates in the coordinate frame of the glasses (may need rotation/translation
                    to represent in world camera's coordinate frame)
    - calibration.xml: info about the camera intrinsics
"""

import shutil
import pathlib
import json
import cv2
import pandas as pd
import numpy as np
import msgpack
import re
import typing
from urllib.request import urlopen

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import timestamps, video_utils


def preprocessData(output_dir: str|pathlib.Path, device: str|EyeTracker=None, source_dir: str|pathlib.Path=None, rec_info: Recording=None, copy_scene_video = True) -> Recording:
    from . import check_folders, check_device
    """
    Run all preprocessing steps on pupil data and store in output_dir
    """
    device, rec_info = check_device(device, rec_info)
    if not device in [EyeTracker.Pupil_Core, EyeTracker.Pupil_Invisible, EyeTracker.Pupil_Neon]:
        raise ValueError(f'Provided device ({rec_info.eye_tracker.value}) is not a {EyeTracker.Pupil_Core.value}, a {EyeTracker.Pupil_Invisible.value} or a {EyeTracker.Pupil_Neon.value}.')
    output_dir, source_dir, rec_info = check_folders(output_dir, source_dir, rec_info, device)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    ### check pupil recording and get export directory
    exportFile, is_cloud_export = checkPupilRecording(source_dir)
    if rec_info is not None:
        checkRecording(source_dir, rec_info)
    else:
        rec_info = getRecordingInfo(source_dir, device)
        if rec_info is None:
            raise RuntimeError(f"The folder {source_dir} is not recognized as a {device.value} recording.")

    # make output dir
    if not output_dir.is_dir():
        output_dir.mkdir()


    # find world video and copy if wanted
    if is_cloud_export:
        scene_vid = list(source_dir.glob('*.mp4'))
        if len(scene_vid)!=1:
            raise RuntimeError(f'Scene video missing or more than one found for Pupil Cloud export in folder {source_dir}')
        srcVid = scene_vid[0]
    else:
        srcVid = source_dir / 'world.mp4'
    if copy_scene_video:
        destVid = output_dir / 'worldCamera.mp4'
        shutil.copy2(str(srcVid), str(destVid))
        rec_info.scene_video_file = destVid.name
    else:
        rec_info.scene_video_file =  srcVid.name
    print(f'  Input data copied to: {output_dir}')


    ### get camera cal
    print('  Getting camera calibration...')
    if is_cloud_export:
        sceneVideoDimensions = getCameraCalFromCloudExport(source_dir, output_dir, rec_info)
    else:
        match rec_info.eye_tracker:
            case EyeTracker.Pupil_Core:
                sceneVideoDimensions = getCameraFromMsgPack(source_dir, output_dir)
            case EyeTracker.Pupil_Invisible:
                if (source_dir/'calibration.bin').is_file():
                    sceneVideoDimensions = getCameraCalFromBinFile(source_dir, output_dir, rec_info)
                else:
                    sceneVideoDimensions = getCameraCalFromOnline(source_dir, output_dir, rec_info)


    ### get gaze data and video frame timestamps
    print('  Prepping gaze data...')
    if is_cloud_export:
        gazeDf, frameTimestamps = formatGazeDataCloudExport(source_dir, exportFile)
    else:
        gazeDf, frameTimestamps = formatGazeDataPupilPlayer(source_dir, exportFile, sceneVideoDimensions, rec_info)

    # write the gaze data to a csv file
    gazeDf.to_csv(str(output_dir / 'gazeData.tsv'), sep='\t', na_rep='nan', float_format="%.8f")

    # also store frame timestamps
    frameTimestamps.to_csv(str(output_dir / 'frameTimestamps.tsv'), sep='\t')

    # store rec info
    rec_info.store_as_json(output_dir)

    return rec_info


def checkPupilRecording(inputDir: str|pathlib.Path):
    """
    This checks that the folder is properly prepared,
    i.e., either:
    - opened in pupil player and an export was run (currently Pupil Core or Pupil Invisible)
    - exported from Pupil Cloud (currently Pupil Invisible or Pupil Neon)
    """
    # check we have an info.player.json file
    if not (inputDir / 'info.player.json').is_file():
        # possibly a pupil cloud export
        if not (inputDir / 'info.json').is_file() or not (inputDir / 'gaze.csv').is_file():
            raise RuntimeError(f'Neither the info.player.json file nor the info.json and gaze.csv files are found for {inputDir}. Either export from Pupil Cloud or, if the folder contains raw sensor data, open the recording in Pupil Player and run an export before importing into glassesValidator.')
        return inputDir / 'gaze.csv', True

    else:
        # check we have an export in the input dir
        inputExpDir = inputDir / 'exports'
        if not inputExpDir.is_dir():
            raise RuntimeError(f'no exports folder for {inputDir}. Perform a recording export in Pupil Player before importing into glassesValidator.')

        # get latest export in that folder that contain a gaze position file
        gpFiles = sorted(list(inputExpDir.rglob('*gaze_positions*.csv')))
        if not gpFiles:
            raise RuntimeError(f'There are no exports in the folder {inputExpDir}. Perform a recording export in Pupil Player before importing into glassesValidator.')

        return gpFiles[-1], False


def getRecordingInfo(inputDir: str|pathlib.Path, device: EyeTracker) -> Recording:
    # returns None if not a recording directory
    recInfo = Recording(source_directory=inputDir, eye_tracker=device)

    if (inputDir / 'info.player.json').is_file():
        # Pupil player export
        match device:
            case EyeTracker.Pupil_Core:
                # check this is not an invisible recording
                file = inputDir / 'info.invisible.json'
                if file.is_file():
                    return None

                file = inputDir / 'info.player.json'
                if not file.is_file():
                    return None
                with open(file, 'r') as j:
                    iInfo = json.load(j)
                recInfo.name = iInfo['recording_name']
                recInfo.start_time = timestamps.Timestamp(int(iInfo['start_time_system_s'])) # UTC in seconds, keep second part
                recInfo.duration   = int(iInfo['duration_s']*1000)                      # in seconds, convert to ms
                recInfo.recording_software_version = iInfo['recording_software_version']

                # get user name, if any
                user_info_file = inputDir / 'user_info.csv'
                if user_info_file.is_file():
                    df = pd.read_csv(user_info_file)
                    nameRow = df['key'].str.contains('name')
                    if any(nameRow):
                        if not pd.isnull(df[nameRow].value).to_numpy()[0]:
                            recInfo.participant = df.loc[nameRow,'value'].to_numpy()[0]

            case EyeTracker.Pupil_Invisible:
                file = inputDir / 'info.invisible.json'
                if not file.is_file():
                    return None
                with open(file, 'r') as j:
                    iInfo = json.load(j)
                recInfo.name = iInfo['template_data']['recording_name']
                recInfo.recording_software_version = iInfo['app_version']
                recInfo.start_time = timestamps.Timestamp(int(iInfo['start_time']//1000000000)) # UTC in nanoseconds, keep second part
                recInfo.duration   = int(iInfo['duration']//1000000)                            # in nanoseconds, convert to ms
                recInfo.glasses_serial = iInfo['glasses_serial_number']
                recInfo.recording_unit_serial = iInfo['android_device_id']
                recInfo.scene_camera_serial = iInfo['scene_camera_serial_number']
                # get participant name
                file = inputDir / 'wearer.json'
                if file.is_file():
                    wearer_id = iInfo['wearer_id']
                    with open(file, 'r') as j:
                        iInfo = json.load(j)
                    if wearer_id==iInfo['uuid']:
                        recInfo.participant = iInfo['name']

            case EyeTracker.Pupil_Neon:
                return None # there are no pupil player exports for the Neon eye tracker

            case _:
                print(f"Device {device} unknown")
                return None
    else:
        # pupil cloud export, for either Pupil Invisible or Pupil Neon
        if device==EyeTracker.Pupil_Core:
            return None

        # raw sensor data also contain an info.json (checked below), so checking
        # that file is not enough to see if this is a Cloud Export. Check gaze.csv
        # presence
        if not (inputDir / 'gaze.csv').is_file():
            return None

        file = inputDir / 'info.json'
        if not file.is_file():
            return None
        with open(file, 'r') as j:
            iInfo = json.load(j)

        # check this is for the expected device
        is_neon = 'Neon' in iInfo['android_device_name'] or 'frame_name' in iInfo
        if device==EyeTracker.Pupil_Invisible and is_neon:
            return None
        elif device==EyeTracker.Pupil_Neon and not is_neon:
            return None

        recInfo.name = iInfo['template_data']['recording_name']
        recInfo.recording_software_version = iInfo['app_version']
        recInfo.start_time = timestamps.Timestamp(int(iInfo['start_time']//1000000000))  # UTC in nanoseconds, keep second part
        recInfo.duration   = int(iInfo['duration']//1000000)                        # in nanoseconds, convert to ms
        if is_neon:
            recInfo.glasses_serial = iInfo['module_serial_number']
        else:
            recInfo.glasses_serial = iInfo['glasses_serial_number']
            recInfo.scene_camera_serial = iInfo['scene_camera_serial_number']
        recInfo.recording_unit_serial = iInfo['android_device_id']
        if is_neon:
            recInfo.firmware_version = f"{iInfo['pipeline_version']} ({iInfo['firmware_version'][0]}.{iInfo['firmware_version'][1]})"
        else:
            recInfo.firmware_version = iInfo['pipeline_version']
        recInfo.participant = iInfo['wearer_name']

    # we got a valid recording and at least some info if we got here
    # return what we've got
    return recInfo


def checkRecording(inputDir: str|pathlib.Path, recInfo: Recording):
    actualRecInfo = getRecordingInfo(inputDir, recInfo.eye_tracker)
    if actualRecInfo is None or recInfo.name!=actualRecInfo.name:
        raise ValueError(f"A recording with the name \"{recInfo.name}\" was not found in the folder {inputDir}.")

    # make sure caller did not mess with recInfo
    if recInfo.duration!=actualRecInfo.duration:
        raise ValueError(f"A recording with the duration \"{recInfo.duration}\" was not found in the folder {inputDir}.")
    if recInfo.start_time.value!=actualRecInfo.start_time.value:
        raise ValueError(f"A recording with the start_time \"{recInfo.start_time.display}\" was not found in the folder {inputDir}.")
    if recInfo.recording_software_version!=actualRecInfo.recording_software_version:
        raise ValueError(f"A recording with the duration \"{recInfo.duration}\" was not found in the folder {inputDir}.")

    # for invisible and neon recordings we have a bit more info
    if recInfo.eye_tracker in [EyeTracker.Pupil_Invisible, EyeTracker.Pupil_Neon]:
        if recInfo.glasses_serial!=actualRecInfo.glasses_serial:
            raise ValueError(f"A recording with the glasses_serial \"{recInfo.glasses_serial}\" was not found in the folder {inputDir}.")
        if recInfo.recording_unit_serial!=actualRecInfo.recording_unit_serial:
            raise ValueError(f"A recording with the recording_unit_serial \"{recInfo.recording_unit_serial}\" was not found in the folder {inputDir}.")
        if recInfo.eye_tracker==EyeTracker.Pupil_Invisible and recInfo.scene_camera_serial!=actualRecInfo.scene_camera_serial:
            raise ValueError(f"A recording with the scene_camera_serial \"{recInfo.scene_camera_serial}\" was not found in the folder {inputDir}.")
        if (recInfo.participant is not None or actualRecInfo.participant is not None) and recInfo.participant!=actualRecInfo.participant:
            raise ValueError(f"A recording with the participant \"{recInfo.participant}\" was not found in the folder {inputDir}.")


def getCameraFromMsgPack(inputDir: str|pathlib.Path, outputDir: str|pathlib.Path) -> list[int]:
    """
    Read camera calibration from recording information file
    """
    camInfo = getCamInfo(inputDir / 'world.intrinsics')

    # rename some fields, ensure they are numpy arrays
    camInfo['cameraMatrix'] = np.array(camInfo.pop('camera_matrix'))
    camInfo['distCoeff']    = np.array(camInfo.pop('dist_coefs')).flatten()
    camInfo['resolution']   = np.array(camInfo['resolution'])

    # store to file
    storeCameraCalibration(camInfo, outputDir)

    return camInfo['resolution']

def getCameraCalFromBinFile(inputDir: str|pathlib.Path, outputDir: str|pathlib.Path, recInfo: Recording) -> list[int]:
    # provided by pupil-labs
    cal = np.fromfile(
        inputDir / 'calibration.bin',
        np.dtype(
            [
                ("serial", "5a"),
                ("scene_camera_matrix", "(3,3)d"),
                ("scene_distortion_coefficients", "8d"),
                ("scene_extrinsics_affine_matrix", "(3,3)d"),
            ]
        ),
    )
    camInfo = {}
    camInfo['serial_number']= str(cal["serial"])
    camInfo['cameraMatrix'] = cal["scene_camera_matrix"].reshape((3,3))
    camInfo['distCoeff']    = cal["scene_distortion_coefficients"].reshape((8,1))
    camInfo['extrinsic']    = cal["scene_extrinsics_affine_matrix"].reshape((3,3))

    # get resolution from the local intrinsics file or scene video
    camInfo['resolution']   = getSceneCameraResolution(inputDir, recInfo)

    # store to xml file
    storeCameraCalibration(camInfo, outputDir)

    return camInfo['resolution']


def getCameraCalFromOnline(inputDir: str|pathlib.Path, outputDir: str|pathlib.Path, recInfo: Recording) -> list[int]:
    """
    Get camera calibration from pupil labs
    """
    url = f'https://api.cloud.pupil-labs.com/v2/hardware/{recInfo.scene_camera_serial}/calibration.v1?json'

    camInfo = json.loads(urlopen(url).read())
    if camInfo['status'] != 'success':
        raise RuntimeError('Camera calibration could not be loaded, response: %s' % (camInfo['message']))

    camInfo = camInfo['result']

    # rename some fields, ensure they are numpy arrays
    camInfo['cameraMatrix'] = np.array(camInfo.pop('camera_matrix'))
    camInfo['distCoeff']    = np.array(camInfo.pop('dist_coefs')).flatten()
    camInfo['rotation']     = np.reshape(np.array(camInfo.pop('rotation_matrix')),(3,3))

    # get resolution from the local intrinsics file or scene video
    camInfo['resolution']   = getSceneCameraResolution(inputDir, recInfo)

    # store to xml file
    storeCameraCalibration(camInfo, outputDir)

    return camInfo['resolution']

def getCameraCalFromCloudExport(inputDir: str|pathlib.Path, outputDir: str|pathlib.Path, recInfo: Recording) -> list[int]:
    file = inputDir / 'scene_camera.json'
    if not file.is_file():
        return None
    with open(file, 'r') as j:
        camInfo = json.load(j)

    camInfo['cameraMatrix'] = np.array(camInfo.pop('camera_matrix'))
    if 'dist_coefs' in camInfo:
        camInfo['distCoeff']    = np.array(camInfo.pop('dist_coefs')).flatten()
    else:
        camInfo['distCoeff']    = np.array(camInfo.pop('distortion_coefficients')).flatten()

    # get resolution from the scene video
    camInfo['resolution']   = getSceneCameraResolution(inputDir, recInfo)

    # store to xml file
    storeCameraCalibration(camInfo, outputDir)

    return camInfo['resolution']

def storeCameraCalibration(camInfo: dict[str, typing.Any], outputDir: str|pathlib.Path):
    fs = cv2.FileStorage(outputDir / 'calibration.xml', cv2.FILE_STORAGE_WRITE)
    for key,value in camInfo.items():
        fs.write(name=key,val=value)
    fs.release()

def getCamInfo(camInfoFile: str|pathlib.Path):
    with open(camInfoFile, 'rb') as f:
        camInfo = msgpack.unpack(f)

    # get keys which denote a camera resolution
    rex = re.compile('^\(\d+, \d+\)$')

    keys = [k for k in camInfo if rex.match(k)]
    if len(keys)!=1:
        raise RuntimeError('No camera intrinsics or intrinsics for more than one camera found')
    return camInfo[keys[0]]

def getSceneCameraResolution(inputDir: str|pathlib.Path, recInfo: Recording):
    if (inputDir / 'world.intrinsics').is_file():
        return np.array(getCamInfo(inputDir / 'world.intrinsics')['resolution'])
    else:
        import cv2
        cap = cv2.VideoCapture(recInfo.get_scene_video_path())
        if cap.isOpened():
            width  = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            cap.release()
        return np.array([width, height])


def formatGazeDataPupilPlayer(inputDir: str|pathlib.Path, exportFile: str|pathlib.Path, sceneVideoDimensions: list[int], recInfo: Recording):
    # convert the json file to pandas dataframe
    df = readGazeDataPupilPlayer(exportFile, sceneVideoDimensions, recInfo)

    # get timestamps for the scene video
    frameTs = video_utils.getFrameTimestampsFromVideo(recInfo.get_scene_video_path())

    # check pupil-labs' frame timestamps because we may need to correct
    # frame indices in case of holes in the video
    # also need this to correctly timestamp gaze samples
    if (inputDir / 'world_lookup.npy').is_file():
        ft = pd.DataFrame(np.load(str(inputDir / 'world_lookup.npy')))
        ft['frame_idx'] = ft.index
        ft.loc[ft['container_idx']==-1,'container_frame_idx'] = -1
        needsAdjust = not ft['frame_idx'].equals(ft['container_frame_idx'])
        # prep for later clean up
        toDrop = [x for x in ft.columns if x not in ['frame_idx','timestamp']]
        # do further adjustment that may be needed
        if needsAdjust:
            # not all video frames were encoded into the video file. Need to adjust
            # frame_idx in the gaze data to match actual video file
            temp = pd.merge(df,ft,on='frame_idx')
            temp['frame_idx'] = temp['container_frame_idx']
            temp = temp.rename(columns={'timestamp_x':'timestamp'})
            toDrop.append('timestamp_y')
            df   = temp.drop(columns=toDrop)
    else:
        ft = pd.DataFrame()
        ft['timestamp'] = np.load(str(inputDir / 'world_timestamps.npy'))*1000.0
        ft.index.name = 'frame_idx'
        # check there are no gaps in the video file
        if df['frame_idx'].max() > ft.index.max():
            raise RuntimeError('It appears there are frames missing in the scene video, but the file world_lookup.npy that would be needed to deal with that is missing. You can generate it by opening the recording in pupil player.')

    # set t=0 to video start time
    t0 = ft['timestamp'].iloc[0]*1000-frameTs['timestamp'].iloc[0]
    df.loc[:,'timestamp'] -= t0

    # set timestamps as index for gaze
    df = df.set_index('timestamp')

    return df, frameTs


def readGazeDataPupilPlayer(file: str|pathlib.Path, sceneVideoDimensions: list[int], recInfo: Recording) -> pd.DataFrame:
    """
    convert the gaze_positions.csv file to a pandas dataframe
    """
    isCore = recInfo.eye_tracker is EyeTracker.Pupil_Core

    df = pd.read_csv(file)

    # drop columns with nothing in them
    df = df.dropna(how='all', axis=1)
    df = df.drop(columns=['base_data'],errors='ignore') # drop these columns if they exist)

    # rename and reorder columns
    lookup = {'gaze_timestamp': 'timestamp',
              'world_index': 'frame_idx',
              'eye_center1_3d_x':'gaze_ori_l_x',
              'eye_center1_3d_y':'gaze_ori_l_y',
              'eye_center1_3d_z':'gaze_ori_l_z',
              'gaze_normal1_x':'gaze_dir_l_x',
              'gaze_normal1_y':'gaze_dir_l_y',
              'gaze_normal1_z':'gaze_dir_l_z',
              'eye_center0_3d_x':'gaze_ori_r_x',   # NB: if monocular setup filming left eye, this is the left eye
              'eye_center0_3d_y':'gaze_ori_r_y',
              'eye_center0_3d_z':'gaze_ori_r_z',
              'gaze_normal0_x':'gaze_dir_r_x',
              'gaze_normal0_y':'gaze_dir_r_y',
              'gaze_normal0_z':'gaze_dir_r_z',
              'norm_pos_x':'gaze_pos_vid_x',
              'norm_pos_y':'gaze_pos_vid_y',
              'gaze_point_3d_x': 'gaze_pos_3d_x',
              'gaze_point_3d_y': 'gaze_pos_3d_y',
              'gaze_point_3d_z': 'gaze_pos_3d_z'}
    df=df.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in df.columns]
    idx.extend([x for x in df.columns if x not in idx])   # append columns not in lookup
    df = df[idx]

    # mark data with insufficient confidence as missing.
    # for pupil core, pupil labs recommends a threshold of 0.6,
    # for the pupil invisible its a binary signal, and
    # confidence 0 should be excluded
    confThresh = 0.6 if isCore else 0
    todo = [x for x in idx if x in lookup.values()]
    toRemove = df.confidence <= confThresh
    for c in todo[2:]:
        df.loc[toRemove,c] = np.nan

    # convert timestamps from s to ms
    df.loc[:,'timestamp'] *= 1000.0

    # turn gaze locations into pixel data with origin in top-left
    df.loc[:,'gaze_pos_vid_x'] *= sceneVideoDimensions[0]
    df.loc[:,'gaze_pos_vid_y'] = (1-df.loc[:,'gaze_pos_vid_y'])*sceneVideoDimensions[1] # turn origin from bottom-left to top-left

    # return the dataframe
    return df

def formatGazeDataCloudExport(inputDir: str|pathlib.Path, exportFile: str|pathlib.Path):
    df = readGazeDataCloudExport(exportFile)

    frameTimestamps = pd.read_csv(inputDir/'world_timestamps.csv')
    frameTimestamps = frameTimestamps.rename(columns={'timestamp [ns]': 'timestamp'})
    frameTimestamps = frameTimestamps.drop(columns=[x for x in frameTimestamps.columns if x not in ['timestamp']])
    frameTimestamps['frame_idx'] = frameTimestamps.index
    frameTimestamps = frameTimestamps.set_index('frame_idx')

    # set t=0 to video start time
    t0_ns = frameTimestamps['timestamp'].iloc[0]
    df.loc[:,'timestamp']               -= t0_ns
    frameTimestamps.loc[:,'timestamp']  -= t0_ns
    df.loc[:,'timestamp']               /= 1000000.0    # convert timestamps from ns to ms
    frameTimestamps.loc[:,'timestamp']  /= 1000000.0

    # set timestamps as index for gaze
    df = df.set_index('timestamp')

    # use the frame timestamps to assign a frame number to each data point
    frameIdx = video_utils.tssToFrameNumber(df.index,frameTimestamps['timestamp'].to_numpy())
    df.insert(0,'frame_idx',frameIdx['frame_idx'])

    return df, frameTimestamps


def readGazeDataCloudExport(file: str|pathlib.Path) -> pd.DataFrame:
    df = pd.read_csv(file)

    # rename and reorder columns
    lookup = {'timestamp [ns]': 'timestamp',
              'gaze x [px]': 'gaze_pos_vid_x',
              'gaze y [px]': 'gaze_pos_vid_y',
              'pupil diameter left [mm]': 'pup_diam_l',
              'pupil diameter right [mm]': 'pup_diam_r',
              'eyeball center left x [mm]': 'gaze_ori_l_x',
              'eyeball center left y [mm]': 'gaze_ori_l_y',
              'eyeball center left z [mm]': 'gaze_ori_l_z',
              'eyeball center right x [mm]': 'gaze_ori_r_x',
              'eyeball center right y [mm]': 'gaze_ori_r_y',
              'eyeball center right z [mm]': 'gaze_ori_r_z',
              'optical axis left x': 'gaze_dir_l_x',
              'optical axis left y': 'gaze_dir_l_y',
              'optical axis left z': 'gaze_dir_l_z',
              'optical axis right x': 'gaze_dir_r_x',
              'optical axis right y': 'gaze_dir_r_y',
              'optical axis right z': 'gaze_dir_r_z',
    }
    df=df.drop(columns=[x for x in df.columns if x not in lookup and x not in ['worn','blink id']])
    df=df.rename(columns=lookup)

    # check if there is an eye states file
    eye_state_file = file.parent / '3d_eye_states.csv'
    if eye_state_file.exists():
        df_eye = pd.read_csv(eye_state_file)
        df_eye = df_eye.drop(columns=[x for x in df_eye.columns if x not in lookup])
        df_eye = df_eye.rename(columns=lookup)
        df = df.join(df_eye.set_index('timestamp'), on='timestamp')

    # mark data where eye tracker is not worn or during blink as missing
    todo = [lookup[k] for k in lookup if lookup[k] in df.columns and lookup[k]!='timestamp']
    toRemove = df.worn == 0
    if 'blink id' in df:
        toRemove = np.logical_or(toRemove, df['blink id']>0)
    for c in todo:
        df.loc[toRemove,c] = np.nan

    # remove last columns we don't need anymore
    df=df.drop(columns=[x for x in df.columns if x in ['worn','blink id']])

    return df


"""
Cast raw SMI ETG data into common format.

The output directory will contain:
    - frameTimestamps.tsv: frame number and corresponding timestamps for each frame in video
    - worldCamera.mp4: the video from the point-of-view scene camera on the glasses
    - gazeData.tsv: gaze data, where all 2D gaze coordinates are represented w/r/t the world camera,
                    and 3D coordinates in the coordinate frame of the glasses (may need rotation/translation
                    to represent in world camera's coordinate frame)
    - calibration.xml: info about the camera intrinsics and transform glasses coordinate system to
                       camera coordinate system
"""

import shutil
import os
import pathlib
import cv2
import pandas as pd
import numpy as np
import math
import configparser
from io import StringIO
from scipy.spatial.transform import Rotation

from ..recording import Recording
from ..eyetracker import EyeTracker
from .. import video_utils


def preprocessData(output_dir: str|pathlib.Path, source_dir: str|pathlib.Path=None, rec_info: Recording=None, copy_scene_video = True) -> Recording:
    from . import check_folders
    """
    Run all preprocessing steps on SMI data and store in output_dir
    """
    # NB: copy_scene_video input argument might be ignored. If ffmpeg is present, it will be used to transcode the scene camera video
    output_dir, source_dir, rec_info = check_folders(output_dir, source_dir, rec_info, EyeTracker.SMI_ETG)
    print(f'processing: {source_dir.name} -> {output_dir}')


    ### check and copy needed files to the output directory
    print('  Check and copy raw data...')
    if rec_info is not None:
        if not checkRecording(source_dir, rec_info, use_return=True):
            raise ValueError(f"A recording with the name \"{rec_info.name}\" was not found in the folder {source_dir}. Check that the name is correct and make sure that you export the scene video and gaze data using BeGaze as described in the glassesValidator manual.")
    else:
        recInfos = getRecordingInfo(source_dir)
        if recInfos is None:
            raise RuntimeError(f"The folder {source_dir} does not contain SMI ETG recordings prepared for glassesValidator. If this is an SMI recording folder, you may not have run the required gaze data and scene video exports yet. See the glassesValidator manual for which exports you should perform with BeGaze first as well as the file naming scheme.")
        rec_info = recInfos[0]  # take first, arbitrarily. If anything else wanted, user should call this function with a correct rec_info themselves

    # make output dirs
    if not output_dir.is_dir():
        output_dir.mkdir()


    ### copy the raw data to the output directory
    checkRecording(source_dir, rec_info)
    copySMIRecordings(source_dir, output_dir, rec_info, copy_scene_video)

    #### prep the data
    print('  Getting camera calibration...')
    sceneVideoDimensions = getCameraFromFile(source_dir, output_dir)
    print('  Prepping gaze data...')
    gazeDf, frameTimestamps = formatGazeData(source_dir, rec_info, sceneVideoDimensions)

    # write the gaze data to a csv file
    gazeDf.to_csv(str(output_dir / 'gazeData.tsv'), sep='\t', na_rep='nan', float_format="%.8f")

    # also store frame timestamps
    frameTimestamps.to_csv(str(output_dir / 'frameTimestamps.tsv'), sep='\t')

    # store rec info
    rec_info.store_as_json(output_dir)

    return rec_info

def getRecordingInfo(inputDir: str|pathlib.Path) -> list[Recording]:
    # returns None if directory does not any recordings

    # NB: can be multiple recordings in an SMI folder

    # but first see this is a recording folder at all
    if not (inputDir / 'codec1.bin').is_file():
        return None
    camInfo = readSMICamInfoFile(inputDir)
    serial = camInfo.get("MiiSimulation",'DeviceSerialNumber')

    # get recordings. We expect the user to rename their exports to have the same format
    # as the other files in a project directory. So e.g., data exported from 001-2-recording.idf
    # for the corresponding 001-2-recording.avi, should be named 001-2-recording.txt. The
    # exported video should be called 001-2-export.avi
    recInfos = []
    for r in inputDir.glob('*-export.avi'):
        recInfos.append(Recording(source_directory=inputDir, eye_tracker=EyeTracker.SMI_ETG))
        recInfos[-1].participant = inputDir.name
        recInfos[-1].name = str(r.name)[:-len('-export.avi')]
        recInfos[-1].glasses_serial = serial

    # should return None if no valid recordings found
    return recInfos if recInfos else None


def checkRecording(inputDir: str|pathlib.Path, recInfo: str|pathlib.Path, use_return = False):
    """
    This checks that the folder is properly prepared
    (i.e. the required BeGaze exports were run)
    """
    # check we have an exported gaze data file
    file = recInfo.name + '-export.avi'
    if not (inputDir / file).is_file():
        if use_return:
            return False
        else:
            raise RuntimeError(f'{file} file not found in {inputDir}. Make sure you export the scene video using BeGaze as described in the glassesValidator manual.')

    # check we have an exported scene video
    file = recInfo.name + '-recording.txt'
    if not (inputDir / file).is_file():
        if use_return:
            return False
        else:
            raise RuntimeError(f'{file} file not found in {inputDir}. Make sure you export the gaze data using BeGaze as described in the glassesValidator manual.')

    return True


def copySMIRecordings(inputDir: pathlib.Path, outputDir: pathlib.Path, recInfo: Recording, copy_scene_video: bool):
    """
    Copy the relevant files from the specified input dir to the specified output dirs
    """

    # Copy relevent files to new directory
    file    = recInfo.name + '-export.avi'
    srcFile = inputDir / file

    # if ffmpeg is on path, remux avi to mp4 (reencode audio from flac to aac as flac is not supported in mp4)
    # else just copy. Ignore copy_scene_video in this case
    if shutil.which('ffmpeg') is not None:
        # make mp4
        destFile = outputDir / 'worldCamera.mp4'
        cmd_str = ' '.join(['ffmpeg', '-hide_banner', '-loglevel', 'error', '-y', '-i', '"'+str(srcFile)+'"', '-vcodec', 'copy', '-acodec', 'aac', '"'+str(destFile)+'"'])
        os.system(cmd_str)
    else:
        if copy_scene_video:
            destFile = outputDir / 'worldCamera.avi'
            shutil.copy2(str(srcFile), str(destFile))
        else:
            destFile = None

    if destFile:
        recInfo.scene_video_file = destFile.name
    else:
        recInfo.scene_video_file =  srcFile.name


def readSMICamInfoFile(inputDir: str|pathlib.Path):
    with open(inputDir / 'codec1.bin', 'r') as file:
        camInfoStr = file.read().replace('## ', '')

    camInfo = configparser.ConfigParser(converters={'nparray': lambda x: np.fromstring(x,sep='\t')})
    camInfo.read_string(camInfoStr)
    return camInfo


def getCameraFromFile(inputDir: str|pathlib.Path, outputDir: str|pathlib.Path):
    """
    Read camera calibration from information file
    """
    camInfo = readSMICamInfoFile(inputDir)

    camera = {}
    camera['FOV'] = camInfo.getfloat("MiiSimulation",'SceneCamFOV')
    camera['resolution'] = np.array([camInfo.getint("MiiSimulation",'SceneCamWidth'), camInfo.getint("MiiSimulation",'SceneCamHeight')])
    camera['sensorOffsets'] = np.array([camInfo.getfloat("MiiSimulation",'SceneCamSensorOffsetX'), camInfo.getfloat("MiiSimulation",'SceneCamSensorOffsetY')])

    camera['radialDistortion'] = camInfo.getnparray("MiiSimulation",'SceneCamRadialDistortion')
    camera['tangentialDistortion'] = camInfo.getnparray("MiiSimulation",'SceneCamTangentialDistortion')

    camera['position'] = camInfo.getnparray("MiiSimulation",'SceneCamPos')
    camera['eulerAngles'] = np.array([camInfo.getfloat("MiiSimulation",'SceneCamOrX'), camInfo.getfloat("MiiSimulation",'SceneCamOrY'), camInfo.getfloat("MiiSimulation",'SceneCamOrZ')])

    # now turn these fields into focal length and principal point
    # 1. FOV is horizontal FOV of camera, given resolution we can compute
    # focal length. We assume vertical focal length is the same, best we can do
    fl = camera['resolution'][0]/(2*math.tan(camera['FOV']/2/180*math.pi))
    camera['focalLength'] = np.array([fl, fl])
    # 2. sensor offsets seem to be relative to center of sensor
    camera['principalPoint'] = camera['resolution']/2.+camera['sensorOffsets']
    # 3. turn euler angles into rotation matrix (180-Rz because poster space has positive X rightward and positive Y downward)
    camera['rotation'] = Rotation.from_euler('XYZ', [camera['eulerAngles'][0], camera['eulerAngles'][1], 180-camera['eulerAngles'][2]], degrees=True).as_matrix()

    # turn into camera matrix and distortion coefficients as used by OpenCV
    camera['cameraMatrix'] = np.identity(3)
    camera['cameraMatrix'][0,0] = camera['focalLength'][0]
    camera['cameraMatrix'][1,1] = camera['focalLength'][1]
    camera['cameraMatrix'][0,2] = camera['principalPoint'][0]
    camera['cameraMatrix'][1,2] = camera['principalPoint'][1]

    camera['distCoeff'] = np.zeros(5)
    camera['distCoeff'][:2]  = camera['radialDistortion'][:2]
    camera['distCoeff'][2:4] = camera['tangentialDistortion']
    camera['distCoeff'][4]   = camera['radialDistortion'][2]


    # store to file
    fs = cv2.FileStorage(outputDir / 'calibration.xml', cv2.FILE_STORAGE_WRITE)
    for key,value in camera.items():
        fs.write(name=key,val=value)
    fs.release()

    return camera['resolution']


def formatGazeData(inputDir: str|pathlib.Path, recInfo: Recording, sceneVideoDimensions: list[int]):
    """
    load gazedata file
    format to get the gaze coordinates w.r.t. world camera, and timestamps for
    every frame of video

    Returns:
        - formatted dataframe with cols for timestamp, frame_idx, and gaze data
        - np array of frame timestamps
    """

    # convert the text file to pandas dataframe
    file = recInfo.name + '-recording.txt'
    df = gazedata2df(inputDir / file, sceneVideoDimensions)

    # read video file, create array of frame timestamps
    frameTimestamps = video_utils.getFrameTimestampsFromVideo(recInfo.get_scene_video_path())

    # SMI frame counter seems to be of the format HH:MM:SS:FR, where HH:MM:SS is a normal
    # hour, minute, second timecode, and FR is a frame number for within that second. The
    # frame number is zero-based, and runs ranges from 0 to FS-1 where FS is the frame
    # rate of the camera (e.g. 24 Hz). Convert their frame counter to a normal counter
    df_fr = pd.DataFrame(df['Frame'].apply(lambda x: [int(y) for y in x.split(':')]).to_list(),columns=['hour','minute','second','frame'])
    frameRate = df_fr.frame.max()+1
    df.insert(0,'frame_idx',(((df_fr.hour*60 + df_fr.minute)*60 + df_fr.second)*frameRate + df_fr.frame).to_numpy())
    # NB: seems we can get frame numbers in the data which are beyond the length of the video. so be it
    df=df.drop(columns=['Frame'])
    # NB: it seems the SMI export doesn't strictly follow their own timecode, but uses the first
    # gaze data point for the first frame of the video, which is then also timestamped with the timecode
    # of that first frame. Subtracting min so that the frame_idx starts at 0 for the data also empirically
    # seems to line up with the SMI export (most of the time, sadly seems to vary a little between videos)
    df.frame_idx = df.frame_idx-df.frame_idx.min()

    # return the gaze data df and frame time stamps array
    return df, frameTimestamps


def gazedata2df(textFile,sceneVideoDimensions):
    """
    convert the gazedata file to a pandas dataframe
    """

    with open(textFile, 'r') as file:
        textData = file.read()

    df = pd.read_table(StringIO(textData),comment='#',index_col=False)

    # get assumed viewing distance
    matchedLines    = [line for line in textData.split('\n') if "Head Distance [mm]" in line]
    viewingDistance = float(matchedLines[0].split('\t')[1])

    # prepare data frame
    # remove unneeded columns
    df=df.drop(columns=['Type', 'Trial', 'Aux1'],errors='ignore') # drop these columns if they exist
    # if we have monocular point of regard data, check if its not identical to binocular
    if 'L POR X [px]' in df.columns:
        xEqual = np.all( np.logical_or(df['L POR X [px]'] == df['B POR X [px]'], np.isnan(df['L POR X [px]']) ))
        yEqual = np.all( np.logical_or(df['L POR Y [px]'] == df['B POR Y [px]'], np.isnan(df['L POR Y [px]']) ))
        if xEqual and yEqual:
            df=df.drop(columns=['L POR X [px]', 'L POR Y [px]'])
    if 'R POR X [px]' in df.columns:
        xEqual = np.all( np.logical_or(df['R POR X [px]'] == df['B POR X [px]'], np.isnan(df['R POR X [px]']) ))
        yEqual = np.all( np.logical_or(df['R POR Y [px]'] == df['B POR Y [px]'], np.isnan(df['R POR Y [px]']) ))
        if xEqual and yEqual:
            df=df.drop(columns=['R POR X [px]', 'R POR Y [px]'])

    # rename and reorder columns
    lookup = {'Time': 'timestamp',
               'L EPOS X':'gaze_ori_l_x',
               'L EPOS Y':'gaze_ori_l_y',
               'L EPOS Z':'gaze_ori_l_z',
               'L Pupil Diameter [mm]':'pup_diam_l',
               'L GVEC X':'gaze_dir_l_x',
               'L GVEC Y':'gaze_dir_l_y',
               'L GVEC Z':'gaze_dir_l_z',
               'R EPOS X':'gaze_ori_r_x',
               'R EPOS Y':'gaze_ori_r_y',
               'R EPOS Z':'gaze_ori_r_z',
               'R Pupil Diameter [mm]':'pup_diam_r',
               'R GVEC X':'gaze_dir_r_x',
               'R GVEC Y':'gaze_dir_r_y',
               'R GVEC Z':'gaze_dir_r_z',
               'B POR X [px]':'gaze_pos_vid_x',
               'B POR Y [px]':'gaze_pos_vid_y',
               'L Dia X [px]':'pup_diam_l_px_x',
               'L Dia Y [px]':'pup_diam_l_px_y',
               'R Dia X [px]':'pup_diam_r_px_x',
               'R Dia Y [px]':'pup_diam_r_px_y'}
    df=df.rename(columns=lookup)
    # reorder
    idx = [lookup[k] for k in lookup if lookup[k] in df.columns]
    idx.extend([x for x in df.columns if x not in idx])   # append columns not in lookup
    df = df[idx]

    # convert timestamps from us to ms and set as index
    df.loc[:,'timestamp'] /= 1000.0
    df = df.set_index('timestamp')

    # return the dataframe
    return df


"""
core.py contains class definitions used by both iso.py and non_iso.py, namely classes (Mp4Box and Mp4FullBox)
that are used as parents for all the real, instantiated boxes. Also contains a header class definition.
"""
import binascii
import logging

from .util import *


class Mp4Box:
    """
    The superclass for all box classes

    """
    def __init__(self, fp, header, parent):
        """ the file pointer, fp will at the same position on exit as entry i.e. at the end of the header"""
        self.header = header
        self.parent = parent
        self.start_of_box = fp.tell() - self.header.header_size
        self.children = []
        self.box_info = {}
        self.byte_string = None
        # only top-level boxes contain an actual byte array for displaying the hex view, lower-level boxes simply
        # take a slice from the top-level box.
        if parent.type == 'file':
            end_of_header = fp.tell()
            fp.seek(self.start_of_box)
            if self.type == 'mdat' and self.size > 1000001:
                self.byte_string = fp.read(1000001)
            else:
                self.byte_string = fp.read(self.size)
            fp.seek(end_of_header)

    @property
    def size(self):
        return self.header.size

    @property
    def type(self):
        return self.header.type

    @type.setter
    def type(self, new_value):
        self.header.type = new_value

    def get_top(self):
        if self.parent.type == 'file':
            return self
        else:
            return self.parent.get_top()

    def get_bytes(self):
        top_box = self.get_top()
        offset = self.start_of_box - top_box.start_of_box
        return top_box.byte_string[offset:offset + self.size]

    def search_child_boxes_for_type(self, box_type):
        type_matches = []
        for box in self.children:
            if box.type == box_type:
                # append object onto array
                type_matches.append(box)
            if box.children:
                # add array onto array
                type_matches += box.search_child_boxes_for_type(box_type)
        return type_matches


class Mp4FullBox(Mp4Box):
    """
    Derived from Mp4Box, but with version and flags.
    """
    def __init__(self, fp, header, parent):
        """ The file pointer, fp will move forward 4 bytes """
        super().__init__(fp, header, parent)
        four_bytes = read_u32(fp)
        self.version = four_bytes >> 24
        self.flags = four_bytes & 0xFFFFFF


class Header:
    """
    All Mp4Boxes contain a header with size and type information.
     """
    def __init__(self, fp):
        """
        The file pointer, fp will be located at the start of the box on entry and at the end of the header on exit
        """
        start_of_box = fp.tell()
        self._size = read_u32(fp)
        my_4bytes = fp.read(4)
        if (struct.unpack('>I', my_4bytes)[0]) >> 24 == 169:
            self.type = my_4bytes[1:].decode('utf-8', errors='ignore')
        else:
            self.type = my_4bytes.decode('utf-8', errors='ignore')
        if self._size == 1:
            self._largesize = read_u64(fp)
        if self.type == 'uuid':
            self.uuid = binascii.b2a_hex(fp.read(16)).decode('utf-8', errors="ignore")
        self.header_size = fp.tell() - start_of_box
        # throw error if size < 8 as 8 bytes is smallest box (free, skip etc)
        if self.size < 8:
            raise Exception('box size {} should be at least 8 bytes.'
                            ' The value of size was: {}'.format(self.type, self.size))

    @property
    def size(self):
        if self._size == 1:
            return self._largesize
        else:
            return self._size

    def get_header(self):
        """ returns all header properties as a dictionary """
        ret_header = {"size": self._size, "type": self.type}
        if self._size == 1:
            ret_header['largesize'] = self._largesize
        if self.type == 'uuid':
            ret_header['uuid'] = self.uuid
        return ret_header



# intervals are either lists of lists ([1,2], [100,200]), or dicts containing such lists as values

def is_in_interval(frame_idx, intervals):
    if intervals is None:
        return True # no interval defined, that means all frames should be processed

    # if its a dict, flatten it
    if isinstance(intervals, dict):
        intervals = [iv for k in intervals for iv in intervals[k]]

    # return True if we're in a current interval
    for iv in intervals:
        if frame_idx>=iv[0] and frame_idx<=iv[1]:
            return True
    return False

def beyond_last_interval(frame_idx, intervals):
    if not intervals:
        return False
    elif isinstance(intervals, dict):
        for k in intervals:
            if intervals[k] and frame_idx <= intervals[k][-1][-1]:
                return False
        return True
    else:
        return frame_idx > intervals[-1][-1]


import numpy as np
import pandas as pd
import pathlib
import importlib
from typing import Any, Optional
from collections import defaultdict


def getColumnLabels(lbl,N=3):
    if N<=3:
        return [lbl+'_%s' % (chr(c)) for c in range(ord('x'), ord('x')+N)]
    elif N==9:
        return [lbl+'[%d,%d]' % (r,c) for r in range(3) for c in range(3)]
    else:
        raise ValueError(f'N input should be <=3 or 9, was {N}')

def noneIfAnyNan(vals):
    if not np.any(np.isnan(vals)):
        return vals
    else:
        return None

def allNanIfNone(vals, numel):
    if vals is None:
        return np.full((numel,), np.nan)
    else:
        return vals


# read coordinate files (e.g. marker files, which have the colums ID, x, y, rotation_angle)
def _read_coord_file_impl(file):
    return pd.read_csv(str(file), index_col='ID')

def read_coord_file(file, package_to_read_from=None):
    # if directory is not provided, try to read the file from the package resources instead
    if package_to_read_from:
        with importlib.resources.path(package_to_read_from, file) as p:
            return _read_coord_file_impl(p)
    else:
        if file.is_file():
            return _read_coord_file_impl(file)
        else:
            return None


def uncompress_columns(cols_compressed: dict[str, int]):
    return [getColumnLabels(c,N) if (N:=cols_compressed[c])>1 else [c] for c in cols_compressed]

def read_file(fileName               : str|pathlib.Path,
              object                 : Any,
              drop_if_all_nan        : bool,
              none_if_any_nan        : bool,
              as_list_dict           : bool,
              make_ori_ts_fridx      : bool,
              episodes               : Optional[list[list[int]]] = None,
              ts_fridx_field_suffixes: list[str]                 = None,
              subset_var                                         = 'frame_idx'):

    # interrogate destination object
    cols_compressed: dict[str, int] = object._columns_compressed
    dtypes         : dict[str, Any] = object._non_float

    # read file and select, if wanted
    df          = pd.read_csv(fileName, delimiter='\t', index_col=False, dtype=defaultdict(lambda: float, **dtypes))
    if episodes:
        sel = (df[subset_var] >= episodes[0][0]) & (df[subset_var] <= episodes[0][1])
        for e in episodes[1:]:
            sel |= (df[subset_var] >= e[0]) & (df[subset_var] <= e[1])
        df = df[sel]

    # figure out what the data columns are
    cols_uncompressed = uncompress_columns(cols_compressed)

    # drop rows where are all data columns are nan
    if drop_if_all_nan:
        df = df.dropna(how='all',subset=[c for cs in cols_uncompressed if len(cs)>1 for c in cs])

    # group columns into numpy arrays, optionally insert None if missing
    for c,ac in zip(cols_compressed,cols_uncompressed):
        if len(ac)==1:
            continue    # nothing to do, would just copy column to itself
        elif ac:
            if not any([a in df.columns for a in ac]):
                continue
            if none_if_any_nan:
                df[c] = [noneIfAnyNan(x) for x in df[ac].values]  # make list of numpy arrays, or None if there are any NaNs in the array
            else:
                df[c] = [             x  for x in df[ac].values]  # make list of numpy arrays
        else:
            df[c] = None

    # keep only the columns we want (this also puts them in the right order even if that doesn't matter since we use kwargs to construct objects)
    df = df[[c for c in cols_compressed.keys() if c in df.columns]]

    # if we have multiple timestamps of frame_idxs, make sure we keep a copy of the original one
    if make_ori_ts_fridx:
        # make copies of original
        df['frame_idx_ori'] = df['frame_idx']
        if 'timestamp' in df.columns:
            df['timestamp_ori'] = df['timestamp']
    # now put requested into normal timestamp column, if wanted
    if make_ori_ts_fridx and ts_fridx_field_suffixes:
        copied = False
        for suf in ts_fridx_field_suffixes: # these are in order of preference
            field = f'frame_idx_{suf}'
            if field not in df.columns:
                continue
            df['frame_idx'] = df[f'frame_idx_{suf}']
            if 'timestamp' in df.columns:
                df['timestamp'] = df[f'timestamp_{suf}']
            copied = True
            break
        assert copied, "None of the specified suffixes were found, can't continue"

    if as_list_dict:
        obj_list = [object(**kwargs) for kwargs in df.to_dict(orient='records')]

        # organize into dict by frame index
        objs = {}
        for k,v in zip(df[subset_var],obj_list):
            objs.setdefault(k, []).append(v)
    else:
        objs = {idx:object(**kwargs) for idx,kwargs in zip(df[subset_var].values,df.to_dict(orient='records'))}
    return objs, df[subset_var].max()

def write_array_to_file(objects             : list[Any] | dict[int,list[Any]],
                        fileName            : str|pathlib.Path,
                        cols_compressed     : dict[str, int],
                        drop_all_nan_cols   : list[str]         = None,
                        skip_all_nan        : bool              = False):
    if not objects:
        return

    if drop_all_nan_cols is None:
        drop_all_nan_cols = []

    if isinstance(objects, dict):
        # flatten
        objects = [o for olist in objects.values() for o in olist]

    records = [{k:getattr(p,k) for k in vars(p) if not k.startswith('_')} for p in objects]
    df = pd.DataFrame.from_records(records)

    # unpack array columns
    cols_uncompressed = uncompress_columns(cols_compressed)
    for c,ac in zip(cols_compressed,cols_uncompressed):
        if len(ac)>1:
            df[ac] = np.vstack([allNanIfNone(v,len(ac)).flatten() for v in df[c].values])

    # if wanted, drop specific columns for which all rows are nan
    if drop_all_nan_cols:
        df = df.drop([c for c in drop_all_nan_cols if c in df and df[c].isnull().all()], axis='columns')

    # if we have filled _ori timestamp and frame_idx columns, copy them back into plain timestamp
    # and frame_idx columns. NB: the _ori columns will be removed because they're not listed in
    # the possible file columns
    if 'timestamp_ori' in df.columns and not df['timestamp_ori'].isnull().all():
        df['timestamp'] = df['timestamp_ori']
    if 'frame_idx_ori' in df.columns and not df['frame_idx_ori'].isnull().all():
        df['frame_idx'] = df['frame_idx_ori']

    # keep only columns to be written out and order them correctly
    df = df[[c for cs in cols_uncompressed for c in cs if c in df.columns]]

    # drop rows where are all data columns are nan
    if skip_all_nan:
        df = df.dropna(how='all', subset=[c for cs in cols_uncompressed if len(cs)>1 for c in cs])

    df.to_csv(fileName, index=False, sep='\t', na_rep='nan', float_format="%.8f")



try:
    from imgui_bundle import imgui, immapp, hello_imgui, glfw_utils
    import glfw
    import OpenGL.GL as gl
except ImportError:
    raise ImportError('imgui_bundle (or one of its dependencies) is not installed, GUI functionality is not available')

import time
import threading
import numpy as np


# simple GUI provider for viewer and coder windows in glassesValidator.process
class GUI:
    def __init__(self, use_thread=True):
        self._should_exit = False
        self._use_thread = use_thread # NB: on MacOSX the GUI needs to be on the main thread, see https://github.com/pthom/hello_imgui/issues/33
        self._thread = None
        self._new_frame = {}
        self._texID = {}
        self._frame_nr = {}
        self._frame_pts = {}
        self._current_frame = {}

        self._next_window_id: int = 0
        self._windows: dict[int,str] = {}
        self._window_flags = int(
                                    imgui.WindowFlags_.no_title_bar |
                                    imgui.WindowFlags_.no_collapse |
                                    imgui.WindowFlags_.no_scrollbar |
                                    imgui.WindowFlags_.no_scroll_with_mouse |
                                    imgui.WindowFlags_.always_auto_resize
                                )
        self._window_visible = {}
        self._window_frame   = {}

        self._interesting_keys = {}
        self._pressed_keys = {}

        self._draw_callback = {'main': None, 'status': None}

    def __del__(self):
        self.stop()

    def add_window(self,name: str) -> int:
        id = self._next_window_id
        self._windows[id] = name
        self._texID[id] = None
        self._new_frame[id] = (None, None, -1)
        self._current_frame[id] = (None, None, -1)
        self._window_visible[id] = False
        self._window_frame[id]   = -1

        self._next_window_id += 1
        return id

    def start(self):
        if not self._windows:
            raise RuntimeError('add a window (GUI.add_window) before you call start')
        if self._use_thread:
            if self._thread is not None:
                raise RuntimeError('The gui is already running, cannot start again')
            self._thread = threading.Thread(target=self._thread_start_fun)
            self._thread.start()
        else:
            self._thread_start_fun()

    def get_state(self):
        return (self._user_closed_window,)

    def stop(self):
        self._should_exit = True
        if self._thread is not None:
            self._thread.join()
        self._thread = None

    def update_image(self, frame, pts, frame_nr, window_id = None):
        # since this has an independently running loop,
        # need to update image whenever new one available
        if window_id is None:
            if len(self._windows)==1:
                window_id = self._get_main_window_id()
            else:
                raise RuntimeError("You have more than one window, you must indicate for which window you are providing an image")

        self._new_frame[window_id] = (frame, pts, frame_nr) # just copy ref to frame is enough

    def register_draw_callback(self, type: str, callback):
        # e.g. for drawing overlays
        if type not in self._draw_callback:
            raise RuntimeError('Draw callback type unknown')
        self._draw_callback[type] = callback

    def set_interesting_keys(self, keys: list[str]):
        if isinstance(keys,str):
            keys = [x for x in keys]

        self._interesting_keys = {}
        self._pressed_keys = {}
        for k in keys:
            # convert to imgui enum
            self._interesting_keys[k] = getattr(imgui.Key,k)
            # storage for presses
            self._pressed_keys[k] = [-1, False]

    def get_key_presses(self) -> dict[str,float]:
        out = {}
        thisT = time.perf_counter()
        for k in self._pressed_keys:
            if self._pressed_keys[k][0] != -1:
                t = thisT - self._pressed_keys[k][0]
                if self._pressed_keys[k][1]:
                    out[k.upper()] = t
                else:
                    out[    k    ] = t

                # clear key press so its not output again
                self._pressed_keys[k] = [-1, False]
        return out

    def _get_main_window_id(self):
        return next(iter(self._windows))

    def _thread_start_fun(self):
        self._lastT=0.
        self._should_exit = False
        self._user_closed_window = False
        self._dpi_fac = 1

        def close_callback(window: glfw._GLFWwindow):
            self._user_closed_window = True

        def post_init():
            imgui.get_io().config_viewports_no_decoration = False
            imgui.get_io().config_viewports_no_auto_merge = True

            glfw.swap_interval(0)
            self._dpi_fac = hello_imgui.dpi_window_size_factor()
            self._window_visible[self._get_main_window_id()] = False
            glfw.set_window_close_callback(glfw_utils.glfw_window_hello_imgui(), close_callback)

        params = hello_imgui.RunnerParams()
        params.app_window_params.window_geometry.size_auto = True
        params.app_window_params.restore_previous_geometry = False
        params.ini_folder_type = hello_imgui.IniFolderType.temp_folder  # so we don't have endless ini files in the app folder, since we don't use them anyway (see previous line, restore_previous_geometry = False)
        params.app_window_params.window_title = self._windows[self._get_main_window_id()]
        params.app_window_params.hidden = True
        params.fps_idling.fps_idle = 0
        params.callbacks.show_gui  = self._gui_func
        params.callbacks.post_init = post_init

        # multiple window support
        params.imgui_window_params.config_windows_move_from_title_bar_only = True
        params.imgui_window_params.enable_viewports = True

        immapp.run(params)

    def _gui_func(self):
        # check if we should exit
        if self._should_exit:
            # clean up
            gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
            for w in self._windows:
                if self._texID[w]:
                    # delete
                    gl.glDeleteTextures(1, [self._texID[w]])
                    self._texID[w] = None
            # and kill
            hello_imgui.get_runner_params().app_shall_exit = True
            # nothing more to do
            return

        # manual vsync with a sleep, so that other thread can run
        # thats crappy vsync, but ok for our purposes
        thisT = time.perf_counter()
        elapsedT = thisT-self._lastT
        self._lastT = thisT

        if elapsedT < 1/60:
            time.sleep(1/60-elapsedT)

        # if user wants to know about keypresses, keep record of them
        for k in self._interesting_keys:
            if imgui.is_key_pressed(self._interesting_keys[k]):
                self._pressed_keys[k] = [thisT, imgui.is_key_down(imgui.Key.im_gui_mod_shift)]

        # upload texture if needed
        for w in self._windows:
            if self._new_frame[w][0] is not None:
                if self._texID[w] is None:
                    self._texID[w] = gl.glGenTextures(1)
                # upload texture
                gl.glBindTexture(gl.GL_TEXTURE_2D, self._texID[w])
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MIN_FILTER, gl.GL_LINEAR)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_MAG_FILTER, gl.GL_LINEAR)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_S, gl.GL_CLAMP_TO_BORDER)
                gl.glTexParameteri(gl.GL_TEXTURE_2D, gl.GL_TEXTURE_WRAP_T, gl.GL_CLAMP_TO_BORDER)
                gl.glPixelStorei(gl.GL_UNPACK_ALIGNMENT, 1)
                gl.glTexImage2D(gl.GL_TEXTURE_2D, 0, gl.GL_RGB, self._new_frame[w][0].shape[1], self._new_frame[w][0].shape[0], 0, gl.GL_BGR, gl.GL_UNSIGNED_BYTE, self._new_frame[w][0])

                # if first time we're showing something
                if self._current_frame[w][0] is None:
                    if w==0:
                        # tell window to resize
                        hello_imgui.get_runner_params().app_window_params.window_geometry.resize_app_window_at_next_frame = True
                        # and show window if needed
                        if not self._window_visible[w]:
                            hello_imgui.get_runner_params().app_window_params.hidden = False
                    # mark window as shown
                    self._window_visible[w] = True

                # keep record of what we're showing
                self._current_frame[w]  = self._new_frame[w]
                self._new_frame[w]      = (None, None, -1)

        # show windows
        for w in self._windows.keys():
            if self._window_visible[w]:
                self._draw_gui(w, w>0)

    def _draw_gui(self, w, need_begin_end):
        if self._current_frame[w] is None or self._texID[w] is None:
            return

        if need_begin_end:
            if isinstance(self._window_size[w][0], np.ndarray):
                imgui.set_next_window_size()
            opened, self._window_visible[w] = imgui.begin(self._windows[w], self._window_visible[w], self._window_flags)
            if not opened:
                imgui.end()
                return

        imgui.set_cursor_pos((0,0))
        # draw image
        self._window_frame[w] += 1
        win_sz = np.array([*imgui.get_window_size()])
        img_sz = np.array([self._current_frame[w][0].shape[1]*self._dpi_fac, self._current_frame[w][0].shape[0]*self._dpi_fac])
        if self._window_frame[w]>2: # it takes two frames for autosize to take effect
            sfac = min([np.min(win_sz/img_sz), 1.]) # clamp to 1: don't scale up
            img_sz = (img_sz * sfac).astype('int')
        imgui.image(self._texID[w], img_sz)

        # draw bottom status overlay
        txt_sz = imgui.calc_text_size('')
        win_bottom = min(self._current_frame[w][0].shape[0]*self._dpi_fac, win_sz[1]+imgui.get_scroll_y())
        imgui.set_cursor_pos_y(win_bottom-txt_sz.y)
        imgui.push_style_color(imgui.Col_.child_bg, (0.0, 0.0, 0.0, 0.6))
        imgui.begin_child("##status_overlay", size=(-imgui.FLT_MIN,txt_sz.y))
        if self._current_frame[w][1] is not None:
            imgui.text(" %8.3f [%d]" % (self._current_frame[w][1], self._current_frame[w][2]))
        else:
            imgui.text(" %d" % (self._current_frame[w][2],))
        if self._draw_callback['status']:
            self._draw_callback['status']()
        imgui.end_child()
        imgui.pop_style_color()

        if self._draw_callback['main']:
            self._draw_callback['main']()

        if need_begin_end:
            imgui.end()

def generic_tooltip_drawer(info_dict: dict[str,str]):
    ws = imgui.get_window_size()
    ts = imgui.calc_text_size('(?)')
    imgui.same_line(ws.x-ts.x)
    imgui.text('(?)')

    if imgui.is_item_hovered():
        imgui.begin_tooltip()
        imgui.push_text_wrap_pos(min(imgui.get_font_size() * 35, ws.x))
        text = ''
        for k in info_dict:
            text += f"'{k.upper()}': {info_dict[k]}\n"
        text = text[:-1]
        imgui.text_unformatted(text)
        imgui.pop_text_wrap_pos()
        imgui.end_tooltip()

def qns_tooltip() -> dict[str,str]:
    return {
        'q': 'Quit',
        'n': 'Next',
        's': 'Screenshot'
    }


import numpy as np
import pandas as pd
import datetime
import warnings
import bisect

from . import utils

class Timestamp:
    def __init__(self, unix_time: int | float, format="%Y-%m-%d %H:%M:%S"):
        self.format = format
        self.display = ""
        self.value = 0
        self.update(unix_time)

    def update(self, unix_time: int | float):
        self.value = int(unix_time)
        if self.value == 0:
            self.display = ""
        else:
            self.display = datetime.datetime.fromtimestamp(unix_time).strftime(self.format)

utils.register_type(utils.CustomTypeEntry(Timestamp,'__Timestamp__',lambda x: x.value, lambda x: Timestamp(x)))


# for reading video timestamp files
class VideoTimestamps:
    def __init__(self, fileName):
        self.timestamp_dict : dict[int,float] = {}
        self.indices        : list[int] = []
        self.timestamps     : list[float] = []
        self._ifi           : float = None

        df = pd.read_csv(fileName, delimiter='\t', index_col='frame_idx')
        self.timestamp_dict = df.to_dict()['timestamp']

        df = df.reset_index()
        df = df[df['frame_idx']!=-1]
        self.indices    = df['frame_idx'].to_list()
        self.timestamps = df['timestamp'].to_list()

    def get_timestamp(self, idx) -> float:
        idx = int(idx)
        if idx in self.timestamp_dict:
            return self.timestamp_dict[idx]
        else:
            warnings.warn('frame_idx %d is not in set\n' % ( idx ), RuntimeWarning )
            return -1.

    def find_frame(self, ts) -> int:
        idx = bisect.bisect(self.timestamps, ts)
        # return nearest
        if idx>=len(self.timestamps):
            return self.indices[-1]
        elif idx>0 and abs(self.timestamps[idx-1]-ts)<abs(self.timestamps[idx]-ts):
            return self.indices[idx-1]
        else:
            return self.indices[idx]

    def get_last(self) -> tuple[int,float]:
        return self.indices[-1], self.timestamps[-1]

    def get_IFI(self) -> float:
        if self._ifi is None:
            self._ifi = np.mean(np.diff(self.timestamps))
        return self._ifi


__version__ = "1.5.1"

__title__ = "glassesTools"
__description__ = "Tools for processing wearable eye tracker recordings."
__url__ = "https://github.com/dcnieho/glassesTools"
__uri__ = __url__
__doc__ = __description__ + " <" + __url__ + ">"

__author__ = "Diederick Niehorster"
__email__ = "diederick_c.niehorster@humlab.lu.se"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2022-2024 " + __author__
